@@ -27,7 +27,11 @@
 
 #include <sys/param.h>
 #include <sys/time.h>
+#ifdef __sun
+#include <sys/sysmacros.h>
+#endif
 
+#include <assert.h>
 #include <ctype.h>
 #include <err.h>
 #include <errno.h>
@@ -53,6 +57,9 @@
 #  define _PATH_DEVNULL "/dev/null"
 #endif
 
+/* Most route(4) messages are less than 256 bytes. */
+#define IOVEC_BUFSIZ	256
+
 #if USE_LOGFILE
 void
 logger_open(struct dhcpcd_ctx *ctx)
@@ -69,7 +76,7 @@ logger_open(struct dhcpcd_ctx *ctx)
 			warn("open: %s", ctx->logfile);
 #ifndef O_CLOEXEC
 		else {
-			if (fcntl(ctx->log_fd, F_GETFD, &f) == -1 ||
+			if ((f = fcntl(ctx->log_fd, F_GETFD)) == -1 ||
 			    fcntl(ctx->log_fd, F_SETFD, f | FD_CLOEXEC) == -1)
 				warn("fcntl: %s", ctx->logfile);
 		}
@@ -276,7 +283,7 @@ addvard(struct dhcpcd_ctx *ctx,
 }
 
 char *
-hwaddr_ntoa(const unsigned char *hwaddr, size_t hwlen, char *buf, size_t buflen)
+hwaddr_ntoa(const uint8_t *hwaddr, size_t hwlen, char *buf, size_t buflen)
 {
 	char *p;
 	size_t i;
@@ -301,16 +308,18 @@ hwaddr_ntoa(const unsigned char *hwaddr, size_t hwlen, char *buf, size_t buflen)
 }
 
 size_t
-hwaddr_aton(unsigned char *buffer, const char *addr)
+hwaddr_aton(uint8_t *buffer, const char *addr)
 {
 	char c[3];
 	const char *p = addr;
-	unsigned char *bp = buffer;
+	uint8_t *bp = buffer;
 	size_t len = 0;
 
 	c[2] = '\0';
 	while (*p) {
 		c[0] = *p++;
+		if (c[0] == '\n')
+			continue;
 		c[1] = *p++;
 		/* Ensure that digits are hex */
 		if (isxdigit((unsigned char)c[0]) == 0 ||
@@ -325,15 +334,98 @@ hwaddr_aton(unsigned char *buffer, const char *addr)
 			return 0;
 		}
 		/* Ensure that next data is EOL or a seperator with data */
-		if (!(*p == '\0' || (*p == ':' && *(p + 1) != '\0'))) {
+		if (!(*p == '\0' || *p == '\n' ||
+		    (*p == ':' && *(p + 1) != '\0')))
+		{
 			errno = EINVAL;
 			return 0;
 		}
 		if (*p)
 			p++;
 		if (bp)
-			*bp++ = (unsigned char)strtol(c, NULL, 16);
+			*bp++ = (uint8_t)strtol(c, NULL, 16);
 		len++;
 	}
 	return len;
 }
+
+size_t
+read_hwaddr_aton(uint8_t **data, const char *path)
+{
+	FILE *fp;
+	char *buf;
+	size_t buf_len, len;
+	ssize_t llen;
+
+	if ((fp = fopen(path, "r")) == NULL)
+		return 0;
+
+	buf = NULL;
+	buf_len = len = 0;
+	*data = NULL;
+	while ((llen = getline(&buf, &buf_len, fp)) != -1) {
+		if ((len = hwaddr_aton(NULL, buf)) != 0) {
+			if (buf_len >= len)
+				*data = (uint8_t *)buf;
+			else {
+				if ((*data = malloc(len)) == NULL)
+					len = 0;
+			}
+			if (len != 0)
+				(void)hwaddr_aton(*data, buf);
+			if (buf_len < len)
+				free(buf);
+			break;
+		}
+	}
+	fclose(fp);
+	return len;
+}
+
+ssize_t
+recvmsg_realloc(int fd, struct msghdr *msg, int flags)
+{
+	struct iovec *iov;
+	ssize_t slen;
+	size_t len;
+	void *n;
+
+	assert(msg != NULL);
+	assert(msg->msg_iov != NULL && msg->msg_iovlen > 0);
+	assert((flags & (MSG_PEEK | MSG_TRUNC)) == 0);
+
+	/* Assume we are reallocing the last iovec. */
+	iov = &msg->msg_iov[msg->msg_iovlen - 1];
+
+	for (;;) {
+		/* Passing MSG_TRUNC should return the actual size needed. */
+		slen = recvmsg(fd, msg, flags | MSG_PEEK | MSG_TRUNC);
+		if (slen == -1)
+			return -1;
+		if (!(msg->msg_flags & MSG_TRUNC))
+			break;
+
+		len = (size_t)slen;
+
+		/* Some kernels return the size of the receive buffer
+		 * on truncation, not the actual size needed.
+		 * So grow the buffer and try again. */
+		if (iov->iov_len == len)
+			len++;
+		else if (iov->iov_len > len)
+			break;
+		len = roundup(len, IOVEC_BUFSIZ);
+		if ((n = realloc(iov->iov_base, len)) == NULL)
+			return -1;
+		iov->iov_base = n;
+		iov->iov_len = len;
+	}
+
+	slen = recvmsg(fd, msg, flags);
+	if (slen != -1 && msg->msg_flags & MSG_TRUNC) {
+		/* This should not be possible ... */
+		errno = ENOBUFS;
+		return -1;
+	}
+	return slen;
+}
