@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -51,6 +51,9 @@
 #define TOSTRING(a)		STRINGIFY(a)
 #define UNUSED(a)		(void)(a)
 
+#define ROUNDUP4(a)		(1 + (((a) - 1) |  3))
+#define ROUNDUP8(a)		(1 + (((a) - 1) |  7))
+
 #define USEC_PER_SEC		1000000L
 #define USEC_PER_NSEC		1000L
 #define NSEC_PER_SEC		1000000000L
@@ -162,6 +165,10 @@ int get_monotonic(struct timespec *);
  * However, this results in a ugly output on the command line
  * and relies on syslogd(8) starting before dhcpcd which is not
  * always the case. */
+#ifdef SMALL
+# undef USE_LOGFILE
+# define USE_LOGFILE 0
+#endif
 #ifndef USE_LOGFILE
 # define USE_LOGFILE 1
 #endif
@@ -190,6 +197,9 @@ ssize_t addvar(struct dhcpcd_ctx *,
 ssize_t addvard(struct dhcpcd_ctx *,
     char ***, const char *, const char *, size_t);
 
-char *hwaddr_ntoa(const unsigned char *, size_t, char *, size_t);
-size_t hwaddr_aton(unsigned char *, const char *);
+char *hwaddr_ntoa(const uint8_t *, size_t, char *, size_t);
+size_t hwaddr_aton(uint8_t *, const char *);
+size_t read_hwaddr_aton(uint8_t **, const char *);
+
+ssize_t recvmsg_realloc(int, struct msghdr *, int);
 #endif
