@@ -64,6 +64,10 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2016 Roy Marples";
 #include "ipv6nd.h"
 #include "script.h"
 
+#ifdef HAVE_UTIL_H
+#include <util.h>
+#endif
+
 #ifdef USE_SIGNALS
 const int dhcpcd_signals[] = {
 	SIGTERM,
@@ -77,34 +81,6 @@ const int dhcpcd_signals[] = {
 const size_t dhcpcd_signals_len = __arraycount(dhcpcd_signals);
 #endif
 
-#if defined(USE_SIGNALS) || !defined(THERE_IS_NO_FORK)
-static pid_t
-read_pid(const char *pidfile)
-{
-	FILE *fp;
-	pid_t pid;
-
-	if ((fp = fopen(pidfile, "r")) == NULL) {
-		errno = ENOENT;
-		return 0;
-	}
-	if (fscanf(fp, "%d", &pid) != 1)
-		pid = 0;
-	fclose(fp);
-	return pid;
-}
-
-static int
-write_pid(int fd, pid_t pid)
-{
-
-	if (ftruncate(fd, (off_t)0) == -1)
-		return -1;
-	lseek(fd, (off_t)0, SEEK_SET);
-	return dprintf(fd, "%d\n", (int)pid);
-}
-#endif
-
 static void
 usage(void)
 {
@@ -226,7 +202,7 @@ dhcpcd_ifafwaiting(const struct interface *ifp)
 {
 	unsigned long long opts;
 
-	if (!ifp->active)
+	if (ifp->active != IF_ACTIVE_USER)
 		return AF_MAX;
 
 	opts = ifp->options->options;
@@ -307,7 +283,7 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 	errno = ENOSYS;
 	return 0;
 #else
-	pid_t pid;
+	pid_t pid, lpid;
 	char buf = '\0';
 	int sidpipe[2], fd;
 
@@ -348,6 +324,9 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 		logger(ctx, LOG_ERR, "fork: %m");
 		return 0;
 	case 0:
+		if ((lpid = pidfile_lock(ctx->pidfile)) != 0)
+			logger(ctx, LOG_ERR, "%s: pidfile_lock %d: %m",
+			    __func__, lpid);
 		setsid();
 		/* Notify parent it's safe to exit as we've detached. */
 		close(sidpipe[0]);
@@ -366,27 +345,19 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 			dup2(fd, STDERR_FILENO);
 			close(fd);
 		}
-		break;
+		ctx->options |= DHCPCD_DAEMONISED;
+		return 0;
 	default:
 		/* Wait for child to detach */
 		close(sidpipe[1]);
 		if (read(sidpipe[0], &buf, 1) == -1)
 			logger(ctx, LOG_ERR, "failed to read child: %m");
 		close(sidpipe[0]);
-		break;
-	}
-	/* Done with the fd now */
-	if (pid != 0) {
 		logger(ctx, LOG_INFO, "forked to background, child pid %d", pid);
-		write_pid(ctx->pid_fd, pid);
-		close(ctx->pid_fd);
-		ctx->pid_fd = -1;
 		ctx->options |= DHCPCD_FORKED;
 		eloop_exit(ctx->eloop, EXIT_SUCCESS);
 		return pid;
 	}
-	ctx->options |= DHCPCD_DAEMONISED;
-	return pid;
 #endif
 }
 
@@ -408,9 +379,6 @@ stop_interface(struct interface *ifp)
 	struct dhcpcd_ctx *ctx;
 
 	ctx = ifp->ctx;
-	if (!ifp->active)
-		goto stop;
-
 	logger(ctx, LOG_INFO, "%s: removing interface", ifp->name);
 	ifp->options->options |= DHCPCD_STOPPING;
 
@@ -424,9 +392,11 @@ stop_interface(struct interface *ifp)
 	eloop_q_timeout_delete(ctx->eloop, 0, NULL, ifp);
 
 	/* De-activate the interface */
-	ifp->active = 0;
+	ifp->active = IF_INACTIVE;
+	ifp->options->options &= ~DHCPCD_STOPPING;
+	/* Set the link state to unknown as we're no longer tracking it. */
+	ifp->carrier = LINK_UNKNOWN;
 
-stop:
 	if (!(ctx->options & (DHCPCD_MASTER | DHCPCD_TEST)))
 		eloop_exit(ctx->eloop, EXIT_FAILURE);
 }
@@ -448,14 +418,17 @@ configure_interface1(struct interface *ifp)
 	if (ifo->options & DHCPCD_RELEASE)
 		ifo->options &= ~DHCPCD_PERSISTENT;
 
-	if (ifp->flags & IFF_POINTOPOINT && !(ifo->options & DHCPCD_INFORM))
-		ifo->options |= DHCPCD_STATIC;
+	if (ifp->flags & (IFF_POINTOPOINT | IFF_LOOPBACK)) {
+		ifo->options &= ~DHCPCD_ARP;
+		if (!(ifp->flags & IFF_MULTICAST))
+			ifo->options &= ~DHCPCD_IPV6RS;
+		if (!(ifo->options & DHCPCD_INFORM))
+			ifo->options |= DHCPCD_STATIC;
+	}
 	if (ifp->flags & IFF_NOARP ||
+	    !(ifo->options & DHCPCD_ARP) ||
 	    ifo->options & (DHCPCD_INFORM | DHCPCD_STATIC))
 		ifo->options &= ~DHCPCD_IPV4LL;
-	if (ifp->flags & (IFF_POINTOPOINT | IFF_LOOPBACK) ||
-	    !(ifp->flags & IFF_MULTICAST))
-		ifo->options &= ~DHCPCD_IPV6RS;
 
 	if (ifo->metric != -1)
 		ifp->metric = (unsigned int)ifo->metric;
@@ -489,25 +462,6 @@ configure_interface1(struct interface *ifp)
 			ifo->options |= DHCPCD_IPV6RA_OWN;
 	}
 
-	/* If we haven't specified a ClientID and our hardware address
-	 * length is greater than DHCP_CHADDR_LEN then we enforce a ClientID
-	 * of the hardware address family and the hardware address.
-	 * If there is no hardware address and no ClientID set,
-	 * force a DUID based ClientID. */
-	if (ifp->hwlen > DHCP_CHADDR_LEN)
-		ifo->options |= DHCPCD_CLIENTID;
-	else if (ifp->hwlen == 0 && !(ifo->options & DHCPCD_CLIENTID))
-		ifo->options |= DHCPCD_CLIENTID | DHCPCD_DUID;
-
-	/* Firewire and InfiniBand interfaces require ClientID and
-	 * the broadcast option being set. */
-	switch (ifp->family) {
-	case ARPHRD_IEEE1394:	/* FALLTHROUGH */
-	case ARPHRD_INFINIBAND:
-		ifo->options |= DHCPCD_CLIENTID | DHCPCD_BROADCAST;
-		break;
-	}
-
 	if (!(ifo->options & DHCPCD_IAID)) {
 		/*
 		 * An IAID is for identifying a unqiue interface within
@@ -602,7 +556,7 @@ dhcpcd_selectprofile(struct interface *ifp, const char *profile)
 	if (ifp->ssid_len) {
 		ssize_t r;
 
-		r = print_string(pssid, sizeof(pssid), ESCSTRING,
+		r = print_string(pssid, sizeof(pssid), OT_ESCSTRING,
 		    ifp->ssid, ifp->ssid_len);
 		if (r == -1) {
 			logger(ifp->ctx, LOG_ERR,
@@ -626,8 +580,11 @@ dhcpcd_selectprofile(struct interface *ifp, const char *profile)
 
 	free_options(ifp->options);
 	ifp->options = ifo;
-	if (profile)
+	if (profile) {
+		add_options(ifp->ctx, ifp->name, ifp->options,
+		    ifp->ctx->argc, ifp->ctx->argv);
 		configure_interface1(ifp);
+	}
 	return 1;
 }
 
@@ -639,12 +596,17 @@ configure_interface(struct interface *ifp, int argc, char **argv,
 
 	old = ifp->options ? ifp->options->mtime : 0;
 	dhcpcd_selectprofile(ifp, NULL);
+	if (ifp->options == NULL) {
+		/* dhcpcd cannot continue with this interface. */
+		ifp->active = IF_INACTIVE;
+		return;
+	}
 	add_options(ifp->ctx, ifp->name, ifp->options, argc, argv);
 	ifp->options->options |= options;
 	configure_interface1(ifp);
 
 	/* If the mtime has changed drop any old lease */
-	if (ifp->options && old != 0 && ifp->options->mtime != old) {
+	if (old != 0 && ifp->options->mtime != old) {
 		logger(ifp->ctx, LOG_WARNING,
 		    "%s: confile file changed, expiring leases", ifp->name);
 		dhcpcd_drop(ifp, 0);
@@ -671,13 +633,21 @@ dhcpcd_pollup(void *arg)
 }
 
 static void
-dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
-    unsigned long long options)
+dhcpcd_initstate2(struct interface *ifp, unsigned long long options)
 {
 	struct if_options *ifo;
 
-	configure_interface(ifp, argc, argv, options);
-	ifo = ifp->options;
+	if (options) {
+		if ((ifo = default_config(ifp->ctx)) == NULL) {
+			logger(ifp->ctx, LOG_ERR, "%s: %s: %m",
+			    ifp->name, __func__);
+			return;
+		}
+		ifo->options |= options;
+		free(ifp->options);
+		ifp->options = ifo;
+	} else
+		ifo = ifp->options;
 
 	if (ifo->options & DHCPCD_IPV4 && ipv4_init(ifp->ctx) == -1) {
 		logger(ifp->ctx, LOG_ERR, "ipv4_init: %m");
@@ -687,15 +657,16 @@ dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
 		logger(ifp->ctx, LOG_ERR, "ipv6_init: %m");
 		ifo->options &= ~DHCPCD_IPV6RS;
 	}
+}
 
-	/* Add our link-local address before upping the interface
-	 * so our RFC7217 address beats the hwaddr based one.
-	 * This needs to happen before PREINIT incase a hook script
-	 * inadvertently ups the interface. */
-	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
-		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
-		ifo->options &= ~DHCPCD_IPV6;
-	}
+static void
+dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
+    unsigned long long options)
+{
+
+	configure_interface(ifp, argc, argv, options);
+	if (ifp->active)
+		dhcpcd_initstate2(ifp, 0);
 }
 
 static void
@@ -772,7 +743,7 @@ dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
 			dhcpcd_handleinterface(ctx, 0, ifp->name);
 #endif
 			if (ifp->wireless) {
-				uint8_t ossid[IF_SSIDSIZE];
+				uint8_t ossid[IF_SSIDLEN];
 #ifdef NOCARRIER_PRESERVE_IP
 				size_t olen;
 
@@ -829,20 +800,6 @@ warn_iaid_conflict(struct interface *ifp, uint8_t *iaid)
 		    ifp->name, ifn->name);
 }
 
-static void
-pre_start(struct interface *ifp)
-{
-
-	/* Add our link-local address before upping the interface
-	 * so our RFC7217 address beats the hwaddr based one.
-	 * This is also a safety check incase it was ripped out
-	 * from under us. */
-	if (ifp->options->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
-		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
-		ifp->options->options &= ~DHCPCD_IPV6;
-	}
-}
-
 void
 dhcpcd_startinterface(void *arg)
 {
@@ -907,7 +864,13 @@ dhcpcd_startinterface(void *arg)
 		}
 	}
 
+	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
+		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		ifo->options &= ~DHCPCD_IPV6;
+	}
 	if (ifo->options & DHCPCD_IPV6) {
+		ipv6_startstatic(ifp);
+
 		if (ifo->options & DHCPCD_IPV6RS)
 			ipv6nd_startrs(ifp);
 
@@ -945,11 +908,13 @@ dhcpcd_startinterface(void *arg)
 		}
 	}
 
+#ifdef INET
 	if (ifo->options & DHCPCD_IPV4) {
 		/* Ensure we have an IPv4 state before starting DHCP */
 		if (ipv4_getstate(ifp) != NULL)
 			dhcp_start(ifp);
 	}
+#endif
 }
 
 static void
@@ -957,7 +922,6 @@ dhcpcd_prestartinterface(void *arg)
 {
 	struct interface *ifp = arg;
 
-	pre_start(ifp);
 	if ((!(ifp->ctx->options & DHCPCD_MASTER) ||
 	    ifp->options->options & DHCPCD_IF_UP) &&
 	    if_up(ifp) == -1)
@@ -985,7 +949,6 @@ static void
 run_preinit(struct interface *ifp)
 {
 
-	pre_start(ifp);
 	if (ifp->ctx->options & DHCPCD_TEST)
 		return;
 
@@ -997,25 +960,30 @@ run_preinit(struct interface *ifp)
 }
 
 void
-dhcpcd_activateinterface(struct interface *ifp)
+dhcpcd_activateinterface(struct interface *ifp, unsigned long long options)
 {
 
 	if (!ifp->active) {
-		ifp->active = 1;
-		dhcpcd_initstate(ifp, 0);
-		run_preinit(ifp);
-		dhcpcd_prestartinterface(ifp);
+		ifp->active = IF_ACTIVE;
+		dhcpcd_initstate2(ifp, options);
+		/* It's possible we might not have been able to load
+		 * a config. */
+		if (ifp->active) {
+			configure_interface1(ifp);
+			run_preinit(ifp);
+			dhcpcd_prestartinterface(ifp);
+		}
 	}
 }
 
 static void
-handle_link(void *arg)
+dhcpcd_handlelink(void *arg)
 {
 	struct dhcpcd_ctx *ctx;
 
 	ctx = arg;
-	if (if_managelink(ctx) == -1) {
-		logger(ctx, LOG_ERR, "if_managelink: %m");
+	if (if_handlelink(ctx) == -1) {
+		logger(ctx, LOG_ERR, "if_handlelink: %m");
 		eloop_event_delete(ctx->eloop, ctx->link_fd);
 		close(ctx->link_fd);
 		ctx->link_fd = -1;
@@ -1060,12 +1028,14 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 			continue;
 
 		/* If running off an interface list, check it's in it. */
-		if (ctx->ifc) {
+		if (ctx->ifc || ctx->options & DHCPCD_INACTIVE) {
 			for (i = 0; i < ctx->ifc; i++)
 				if (strcmp(ctx->ifv[i], ifname) == 0)
 					break;
-			if (i >= ctx->ifc)
-				ifp->active = 0;
+			if (i >= ctx->ifc) {
+				ifp->active = IF_INACTIVE;
+				ifp->carrier = LINK_UNKNOWN;
+			}
 		}
 
 		i = 0;
@@ -1109,7 +1079,7 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 
 void
 dhcpcd_handlehwaddr(struct dhcpcd_ctx *ctx, const char *ifname,
-    const uint8_t *hwaddr, uint8_t hwlen)
+    const void *hwaddr, uint8_t hwlen)
 {
 	struct interface *ifp;
 	char buf[sizeof(ifp->hwaddr) * 3];
@@ -1152,7 +1122,8 @@ reload_config(struct dhcpcd_ctx *ctx)
 	struct if_options *ifo;
 
 	free_globals(ctx);
-	ifo = read_config(ctx, NULL, NULL, NULL);
+	if ((ifo = read_config(ctx, NULL, NULL, NULL)) == NULL)
+		return;
 	add_options(ctx, NULL, ifo, ctx->argc, ctx->argv);
 	/* We need to preserve these two options. */
 	if (ctx->options & DHCPCD_MASTER)
@@ -1182,7 +1153,7 @@ reconf_reboot(struct dhcpcd_ctx *ctx, int action, int argc, char **argv, int oi)
 			else
 				ipv4_applyaddr(ifp);
 		} else if (i != argc) {
-			ifp->active = 1;
+			ifp->active = IF_ACTIVE_USER;
 			dhcpcd_initstate1(ifp, argc, argv, 0);
 			run_preinit(ifp);
 			dhcpcd_prestartinterface(ifp);
@@ -1195,15 +1166,16 @@ stop_all_interfaces(struct dhcpcd_ctx *ctx, unsigned long long opts)
 {
 	struct interface *ifp;
 
+	ctx->options |= DHCPCD_EXITING;
 	/* Drop the last interface first */
 	TAILQ_FOREACH_REVERSE(ifp, ctx->ifaces, if_head, next) {
-		if (ifp->options) {
+		if (ifp->active) {
 			ifp->options->options |= opts;
 			if (ifp->options->options & DHCPCD_RELEASE)
 				ifp->options->options &= ~DHCPCD_PERSISTENT;
 			ifp->options->options |= DHCPCD_EXITING;
+			stop_interface(ifp);
 		}
-		stop_interface(ifp);
 	}
 }
 
@@ -1306,6 +1278,8 @@ dhcpcd_getinterfaces(void *arg)
 			len++;
 		if (IPV4LL_STATE_RUNNING(ifp))
 			len++;
+		if (IPV6_STATE_RUNNING(ifp))
+			len++;
 		if (RS_STATE_RUNNING(ifp))
 			len++;
 		if (D6_STATE_RUNNING(ifp))
@@ -1344,7 +1318,7 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
 		return control_queue(fd, UNCONST(fd->ctx->cffile),
 		    strlen(fd->ctx->cffile) + 1, 0);
 	} else if (strcmp(*argv, "--getinterfaces") == 0) {
-		eloop_event_add(fd->ctx->eloop, fd->fd, NULL, NULL,
+		eloop_event_add_w(fd->ctx->eloop, fd->fd,
 		    dhcpcd_getinterfaces, fd);
 		return 0;
 	} else if (strcmp(*argv, "--listen") == 0) {
@@ -1414,6 +1388,8 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
 		for (oi = optind; oi < argc; oi++) {
 			if ((ifp = if_find(ctx->ifaces, argv[oi])) == NULL)
 				continue;
+			if (!ifp->active)
+				continue;
 			ifp->options->options |= opts;
 			if (opts & DHCPCD_RELEASE)
 				ifp->options->options &= ~DHCPCD_PERSISTENT;
@@ -1466,12 +1442,28 @@ main(int argc, char **argv)
 			return EXIT_SUCCESS;
 		} else if (strcmp(argv[1], "--version") == 0) {
 			printf(""PACKAGE" "VERSION"\n%s\n", dhcpcd_copyright);
+			printf("Compiled in features:"
+#ifdef INET
+			" INET"
+#endif
+#ifdef IPV4LL
+			" IPv4LL"
+#endif
+#ifdef INET6
+			" INET6"
+#endif
+#ifdef DHCP6
+			" DHCPv6"
+#endif
+#ifdef AUTH
+			" AUTH"
+#endif
+			"\n");
 			return EXIT_SUCCESS;
 		}
 	}
 
 	memset(&ctx, 0, sizeof(ctx));
-	closefrom(3);
 
 	ctx.log_fd = -1;
 	logger_open(&ctx);
@@ -1479,11 +1471,8 @@ main(int argc, char **argv)
 
 	ifo = NULL;
 	ctx.cffile = CONFIG;
-	ctx.pid_fd = ctx.control_fd = ctx.control_unpriv_fd = ctx.link_fd = -1;
+	ctx.control_fd = ctx.control_unpriv_fd = ctx.link_fd = -1;
 	ctx.pf_inet_fd = -1;
-#if defined(INET6) && defined(BSD)
-	ctx.pf_inet6_fd = -1;
-#endif
 #ifdef IFLR_ACTIVE
 	ctx.pf_link_fd = -1;
 #endif
@@ -1705,7 +1694,8 @@ printpidfile:
 			ifp->ctx = &ctx;
 			TAILQ_INSERT_HEAD(ctx.ifaces, ifp, next);
 			if (family == 0) {
-				if (ctx.pidfile[strlen(ctx.pidfile) - 1] == '6')
+				if (ctx.pidfile[0] != '\0' &&
+				    ctx.pidfile[strlen(ctx.pidfile) - 1] == '6')
 					family = AF_INET6;
 				else
 					family = AF_INET;
@@ -1734,13 +1724,11 @@ printpidfile:
 	    !(ctx.options & DHCPCD_TEST))
 	{
 #endif
-		if (ctx.options & DHCPCD_MASTER)
-			i = -1;
-		else
-			i = control_open(&ctx, argv[optind]);
-		if (i == -1)
-			i = control_open(&ctx, NULL);
-		if (i != -1) {
+		if (!(ctx.options & DHCPCD_MASTER))
+			ctx.control_fd = control_open(argv[optind]);
+		if (ctx.control_fd == -1)
+			ctx.control_fd = control_open(NULL);
+		if (ctx.control_fd != -1) {
 			logger(&ctx, LOG_INFO,
 			    "sending commands to master dhcpcd process");
 			len = control_send(&ctx, argc, argv);
@@ -1763,14 +1751,14 @@ printpidfile:
 
 #ifdef USE_SIGNALS
 	if (sig != 0) {
-		pid = read_pid(ctx.pidfile);
-		if (pid != 0)
+		pid = pidfile_read(ctx.pidfile);
+		if (pid != 0 && pid != -1)
 			logger(&ctx, LOG_INFO, "sending signal %s to pid %d",
 			    siga, pid);
-		if (pid == 0 || kill(pid, sig) != 0) {
+		if (pid == 0 || pid == -1 || kill(pid, sig) != 0) {
 			if (sig != SIGHUP && sig != SIGUSR1 && errno != EPERM)
 				logger(&ctx, LOG_ERR, ""PACKAGE" not running");
-			if (pid != 0 && errno != ESRCH) {
+			if (pid != 0 && pid != -1 && errno != ESRCH) {
 				logger(&ctx, LOG_ERR, "kill: %m");
 				goto exit_failure;
 			}
@@ -1789,7 +1777,7 @@ printpidfile:
 			ts.tv_nsec = 100000000; /* 10th of a second */
 			for(i = 0; i < 100; i++) {
 				nanosleep(&ts, NULL);
-				if (read_pid(ctx.pidfile) == 0)
+				if (pidfile_read(ctx.pidfile) == -1)
 					goto exit_success;
 			}
 			logger(&ctx, LOG_ERR, "pid %d failed to exit", pid);
@@ -1798,53 +1786,21 @@ printpidfile:
 	}
 
 	if (!(ctx.options & DHCPCD_TEST)) {
-		if ((pid = read_pid(ctx.pidfile)) > 0 &&
-		    kill(pid, 0) == 0)
-		{
-			logger(&ctx, LOG_ERR, ""PACKAGE
-			    " already running on pid %d (%s)",
-			    pid, ctx.pidfile);
-			goto exit_failure;
-		}
-
 		/* Ensure we have the needed directories */
 		if (mkdir(RUNDIR, 0755) == -1 && errno != EEXIST)
 			logger(&ctx, LOG_ERR, "mkdir `%s': %m", RUNDIR);
 		if (mkdir(DBDIR, 0755) == -1 && errno != EEXIST)
 			logger(&ctx, LOG_ERR, "mkdir `%s': %m", DBDIR);
 
-		opt = O_WRONLY | O_CREAT | O_NONBLOCK;
-#ifdef O_CLOEXEC
-		opt |= O_CLOEXEC;
-#endif
-		ctx.pid_fd = open(ctx.pidfile, opt, 0664);
-		if (ctx.pid_fd == -1)
-			logger(&ctx, LOG_ERR, "open `%s': %m", ctx.pidfile);
-		else {
-#ifdef LOCK_EX
-			/* Lock the file so that only one instance of dhcpcd
-			 * runs on an interface */
-			if (flock(ctx.pid_fd, LOCK_EX | LOCK_NB) == -1) {
-				logger(&ctx, LOG_ERR, "flock `%s': %m",
-				    ctx.pidfile);
-				/* We don't want to unlink the pidfile as
-				 * another dhcpcd instance could be using it. */
-				ctx.pidfile[0] = '\0';
-				goto exit_failure;
-			}
-#endif
-#ifndef O_CLOEXEC
-			if (fcntl(ctx.pid_fd, F_GETFD, &opt) == -1 ||
-			    fcntl(ctx.pid_fd, F_SETFD, opt | FD_CLOEXEC) == -1)
-			{
-				logger(&ctx, LOG_ERR, "fcntl: %m");
-				/* We don't want to unlink the pidfile as
-				 * another dhcpcd instance could be using it. */
-				ctx.pidfile[0] = '\0';
-				goto exit_failure;
-			}
-#endif
-			write_pid(ctx.pid_fd, getpid());
+		if ((pid = pidfile_lock(ctx.pidfile)) != 0) {
+			if (pid == -1)
+				logger(&ctx, LOG_ERR, "%s: pidfile_lock: %m",
+				    __func__);
+			else
+				logger(&ctx, LOG_ERR, ""PACKAGE
+				    " already running on pid %d (%s)",
+				    pid, ctx.pidfile);
+			goto exit_failure;
 		}
 	}
 
@@ -1885,7 +1841,7 @@ printpidfile:
 
 	/* Start handling kernel messages for interfaces, addreses and
 	 * routes. */
-	eloop_event_add(ctx.eloop, ctx.link_fd, handle_link, &ctx, NULL, NULL);
+	eloop_event_add(ctx.eloop, ctx.link_fd, dhcpcd_handlelink, &ctx);
 
 	/* Start any dev listening plugin which may want to
 	 * change the interface name provided by the kernel */
@@ -1906,12 +1862,14 @@ printpidfile:
 			    ctx.ifv[i]);
 	}
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		if (ifp->active)
+		if (ifp->active == IF_ACTIVE_USER)
 			break;
 	}
 	if (ifp == NULL) {
 		if (ctx.ifc == 0)
-			logger(&ctx, LOG_ERR, "no valid interfaces found");
+			logger(&ctx,
+			    ctx.options & DHCPCD_INACTIVE ? LOG_DEBUG : LOG_ERR,
+			    "no valid interfaces found");
 		else
 			goto exit_failure;
 		if (!(ctx.options & DHCPCD_LINK)) {
@@ -1955,7 +1913,9 @@ printpidfile:
 		    ctx.options & DHCPCD_LINK &&
 		    !(ctx.options & DHCPCD_WAITIP))
 		{
-			logger(&ctx, LOG_WARNING,
+			logger(&ctx,
+			    ctx.options & DHCPCD_INACTIVE ?
+			    LOG_DEBUG : LOG_WARNING,
 			    "no interfaces have a carrier");
 			if (dhcpcd_daemonise(&ctx))
 				goto exit_success;
@@ -2005,17 +1965,7 @@ exit1:
 		eloop_event_delete(ctx.eloop, ctx.link_fd);
 		close(ctx.link_fd);
 	}
-	if (ctx.pf_inet_fd != -1)
-		close(ctx.pf_inet_fd);
-#if defined(INET6) && defined(BSD)
-	if (ctx.pf_inet6_fd != -1)
-		close(ctx.pf_inet6_fd);
-#endif
-#ifdef IFLR_ACTIVE
-	if (ctx.pf_link_fd != -1)
-		close(ctx.pf_link_fd);
-#endif
-
+	if_closesockets(&ctx);
 	free_options(ifo);
 	free_globals(&ctx);
 	ipv4_ctxfree(&ctx);
@@ -2023,16 +1973,16 @@ exit1:
 	dev_stop(&ctx);
 	if (control_stop(&ctx) == -1)
 		logger(&ctx, LOG_ERR, "control_stop: %m:");
-	if (ctx.pid_fd != -1) {
-		close(ctx.pid_fd);
-		if (ctx.pidfile[0] != '\0')
-			unlink(ctx.pidfile);
-	}
 	eloop_free(ctx.eloop);
+	free(ctx.iov[0].iov_base);
 
 	if (ctx.options & DHCPCD_STARTED && !(ctx.options & DHCPCD_FORKED))
 		logger(&ctx, LOG_INFO, PACKAGE " exited");
 	logger_close(&ctx);
 	free(ctx.logfile);
+#ifdef USE_SIGNALS
+	if (ctx.options & DHCPCD_FORKED)
+		_exit(i); /* so atexit won't remove our pidfile */
+#endif
 	return i;
 }
