@@ -1,7 +1,7 @@
 /*
- * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
- * All rights reserved
+ * eloop - portable event based main loop.
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
+ * All rights reserved.
 
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -25,18 +25,59 @@
  * SUCH DAMAGE.
  */
 
+#if (defined(__unix__) || defined(unix)) && !defined(USG)
+#include <sys/param.h>
+#endif
 #include <sys/time.h>
 
 #include <assert.h>
 #include <errno.h>
 #include <limits.h>
 #include <signal.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
-/* config.h should define HAVE_KQUEUE, HAVE_EPOLL, etc */
+/* config.h should define HAVE_KQUEUE, HAVE_EPOLL, etc. */
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
+
+/* Attempt to autodetect kqueue or epoll.
+ * Failing that, fall back to pselect. */
+#if !defined(HAVE_KQUEUE) && !defined(HAVE_EPOLL) && !defined(HAVE_PSELECT) && \
+    !defined(HAVE_POLLTS) && !defined(HAVE_PPOLL)
+#if defined(BSD)
+/* Assume BSD has a working sys/queue.h and kqueue(2) interface. */
+#define HAVE_SYS_QUEUE_H
+#define HAVE_KQUEUE
+#elif defined(__linux__)
+/* Assume Linux has a working epoll(3) interface. */
+#define HAVE_EPOLL
+#else
+/* pselect(2) is a POSIX standard. */
+#define HAVE_PSELECT
+#endif
+#endif
+
+/* pollts and ppoll require poll.
+ * pselect is wrapped in a pollts/ppoll style interface
+ * and as such require poll as well. */
+#if defined(HAVE_PSELECT) || defined(HAVE_POLLTS) || defined(HAVE_PPOLL)
+#ifndef HAVE_POLL
+#define HAVE_POLL
+#endif
+#if defined(HAVE_POLLTS)
+#define POLLTS pollts
+#elif defined(HAVE_PPOLL)
+#define POLLTS ppoll
+#else
+#define POLLTS eloop_pollts
+#define ELOOP_NEED_POLLTS
+#endif
+#endif
+
 #include "eloop.h"
 
 #ifndef UNUSED
@@ -59,20 +100,106 @@
 #include <sys/event.h>
 #include <fcntl.h>
 #ifdef __NetBSD__
-/* udata is void * except on NetBSD
- * lengths are int except on NetBSD */
+/* udata is void * except on NetBSD.
+ * lengths are int except on NetBSD. */
 #define UPTR(x)	((intptr_t)(x))
 #define LENC(x)	(x)
 #else
 #define UPTR(x)	(x)
 #define LENC(x)	((int)(x))
 #endif
-#define eloop_event_setup_fds(eloop)
 #elif defined(HAVE_EPOLL)
 #include <sys/epoll.h>
-#define eloop_event_setup_fds(eloop)
-#else
+#elif defined(HAVE_POLL)
+#if defined(HAVE_PSELECT)
+#include <sys/select.h>
+#endif
 #include <poll.h>
+#endif
+
+/* Our structures require TAILQ macros, which really every libc should
+ * ship as they are useful beyond belief.
+ * Sadly some libc's don't have sys/queue.h and some that do don't have
+ * the TAILQ_FOREACH macro. For those that don't, the application using
+ * this implementation will need to ship a working queue.h somewhere.
+ * If we don't have sys/queue.h found in config.h, then
+ * allow QUEUE_H to override loading queue.h in the current directory. */
+#ifndef TAILQ_FOREACH
+#ifdef HAVE_SYS_QUEUE_H
+#include <sys/queue.h>
+#elif defined(QUEUE_H)
+#define __QUEUE_HEADER(x) #x
+#define _QUEUE_HEADER(x) __QUEUE_HEADER(x)
+#include _QUEUE_HEADER(QUEUE_H)
+#else
+#include "queue.h"
+#endif
+#endif
+
+struct eloop_event {
+	TAILQ_ENTRY(eloop_event) next;
+	int fd;
+	void (*read_cb)(void *);
+	void *read_cb_arg;
+	void (*write_cb)(void *);
+	void *write_cb_arg;
+};
+
+struct eloop_timeout {
+	TAILQ_ENTRY(eloop_timeout) next;
+	struct timespec when;
+	void (*callback)(void *);
+	void *arg;
+	int queue;
+};
+
+struct eloop {
+	size_t events_len;
+	TAILQ_HEAD (event_head, eloop_event) events;
+	struct event_head free_events;
+	int events_maxfd;
+	struct eloop_event **event_fds;
+
+	TAILQ_HEAD (timeout_head, eloop_timeout) timeouts;
+	struct timeout_head free_timeouts;
+
+	void (*timeout0)(void *);
+	void *timeout0_arg;
+	const int *signals;
+	size_t signals_len;
+	void (*signal_cb)(int, void *);
+	void *signal_cb_ctx;
+
+#if defined(HAVE_KQUEUE) || defined(HAVE_EPOLL)
+	int poll_fd;
+#elif defined(HAVE_POLL)
+	struct pollfd *fds;
+	size_t fds_len;
+#endif
+
+	int exitnow;
+	int exitcode;
+};
+
+#ifdef HAVE_REALLOCARRAY
+#define	eloop_realloca	reallocarray
+#else
+/* Handy routing to check for potential overflow.
+ * reallocarray(3) and reallocarr(3) are not portable. */
+#define SQRT_SIZE_MAX (((size_t)1) << (sizeof(size_t) * CHAR_BIT / 2))
+static void *
+eloop_realloca(void *ptr, size_t n, size_t size)
+{
+
+	if ((n | size) >= SQRT_SIZE_MAX && n > SIZE_MAX / size) {
+		errno = EOVERFLOW;
+		return NULL;
+	}
+	return realloc(ptr, n * size);
+}
+#endif
+
+#ifdef HAVE_POLL
 static void
 eloop_event_setup_fds(struct eloop *eloop)
 {
@@ -88,22 +215,22 @@ eloop_event_setup_fds(struct eloop *eloop)
 		if (e->write_cb)
 			eloop->fds[i].events |= POLLOUT;
 		eloop->fds[i].revents = 0;
-		e->pollfd = &eloop->fds[i];
 		i++;
 	}
 }
 
-#ifndef pollts
+#ifdef ELOOP_NEED_POLLTS
 /* Wrapper around pselect, to imitate the NetBSD pollts call. */
 static int
-pollts(struct pollfd * fds, nfds_t nfds,
+eloop_pollts(struct pollfd * fds, nfds_t nfds,
     const struct timespec *ts, const sigset_t *sigmask)
 {
-	fd_set read_fds;
+	fd_set read_fds, write_fds;
 	nfds_t n;
 	int maxfd, r;
 
 	FD_ZERO(&read_fds);
+	FD_ZERO(&write_fds);
 	maxfd = 0;
 	for (n = 0; n < nfds; n++) {
 		if (fds[n].events & POLLIN) {
@@ -111,23 +238,32 @@ pollts(struct pollfd * fds, nfds_t nfds,
 			if (fds[n].fd > maxfd)
 				maxfd = fds[n].fd;
 		}
+		if (fds[n].events & POLLOUT) {
+			FD_SET(fds[n].fd, &write_fds);
+			if (fds[n].fd > maxfd)
+				maxfd = fds[n].fd;
+		}
 	}
 
-	r = pselect(maxfd + 1, &read_fds, NULL, NULL, ts, sigmask);
+	r = pselect(maxfd + 1, &read_fds, &write_fds, NULL, ts, sigmask);
 	if (r > 0) {
 		for (n = 0; n < nfds; n++) {
 			fds[n].revents =
 			    FD_ISSET(fds[n].fd, &read_fds) ? POLLIN : 0;
+			if (FD_ISSET(fds[n].fd, &write_fds))
+				fds[n].revents |= POLLOUT;
 		}
 	}
 
 	return r;
 }
-#endif
-#endif
+#endif /* pollts */
+#else /* !HAVE_POLL */
+#define eloop_event_setup_fds(a) {}
+#endif /* HAVE_POLL */
 
 int
-eloop_event_add(struct eloop *eloop, int fd,
+eloop_event_add_rw(struct eloop *eloop, int fd,
     void (*read_cb)(void *), void *read_cb_arg,
     void (*write_cb)(void *), void *write_cb_arg)
 {
@@ -136,7 +272,7 @@ eloop_event_add(struct eloop *eloop, int fd,
 	struct kevent ke[2];
 #elif defined(HAVE_EPOLL)
 	struct epoll_event epe;
-#else
+#elif defined(HAVE_POLL)
 	struct pollfd *nfds;
 #endif
 
@@ -155,9 +291,9 @@ eloop_event_add(struct eloop *eloop, int fd,
 		epe.events |= EPOLLOUT;
 #endif
 
-	/* We should only have one callback monitoring the fd */
-	TAILQ_FOREACH(e, &eloop->events, next) {
-		if (e->fd == fd) {
+	/* We should only have one callback monitoring the fd. */
+	if (fd <= eloop->events_maxfd) {
+		if ((e = eloop->event_fds[fd]) != NULL) {
 			int error;
 
 #if defined(HAVE_KQUEUE)
@@ -189,9 +325,26 @@ eloop_event_add(struct eloop *eloop, int fd,
 			eloop_event_setup_fds(eloop);
 			return error;
 		}
+	} else {
+		struct eloop_event **new_fds;
+		int maxfd, i;
+
+		/* Reserve ourself and 4 more. */
+		maxfd = fd + 4;
+		new_fds = eloop_realloca(eloop->event_fds,
+		    ((size_t)maxfd + 1), sizeof(*eloop->event_fds));
+		if (new_fds == NULL)
+			return -1;
+
+		/* set new entries NULL as the fd's may not be contiguous. */
+		for (i = maxfd; i > eloop->events_maxfd; i--)
+			new_fds[i] = NULL;
+
+		eloop->event_fds = new_fds;
+		eloop->events_maxfd = maxfd;
 	}
 
-	/* Allocate a new event if no free ones already allocated */
+	/* Allocate a new event if no free ones already allocated. */
 	if ((e = TAILQ_FIRST(&eloop->free_events))) {
 		TAILQ_REMOVE(&eloop->free_events, e, next);
 	} else {
@@ -200,12 +353,12 @@ eloop_event_add(struct eloop *eloop, int fd,
 			goto err;
 	}
 
-	/* Ensure we can actually listen to it */
+	/* Ensure we can actually listen to it. */
 	eloop->events_len++;
-#if !defined(HAVE_KQUEUE) && !defined(HAVE_EPOLL)
+#ifdef HAVE_POLL
 	if (eloop->events_len > eloop->fds_len) {
-		nfds = realloc(eloop->fds,
-		    sizeof(*eloop->fds) * (eloop->fds_len + 5));
+		nfds = eloop_realloca(eloop->fds,
+		    (eloop->fds_len + 5), sizeof(*eloop->fds));
 		if (nfds == NULL)
 			goto err;
 		eloop->fds_len += 5;
@@ -213,7 +366,7 @@ eloop_event_add(struct eloop *eloop, int fd,
 	}
 #endif
 
-	/* Now populate the structure and add it to the list */
+	/* Now populate the structure and add it to the list. */
 	e->fd = fd;
 	e->read_cb = read_cb;
 	e->read_cb_arg = read_cb_arg;
@@ -235,13 +388,8 @@ eloop_event_add(struct eloop *eloop, int fd,
 		goto err;
 #endif
 
-	/* The order of events should not matter.
-	 * However, some PPP servers love to close the link right after
-	 * sending their final message. So to ensure dhcpcd processes this
-	 * message (which is likely to be that the DHCP addresses are wrong)
-	 * we insert new events at the queue head as the link fd will be
-	 * the first event added. */
 	TAILQ_INSERT_HEAD(&eloop->events, e, next);
+	eloop->event_fds[e->fd] = e;
 	eloop_event_setup_fds(eloop);
 	return 0;
 
@@ -253,6 +401,22 @@ err:
 	return -1;
 }
 
+int
+eloop_event_add(struct eloop *eloop, int fd,
+    void (*read_cb)(void *), void *read_cb_arg)
+{
+
+	return eloop_event_add_rw(eloop, fd, read_cb, read_cb_arg, NULL, NULL);
+}
+
+int
+eloop_event_add_w(struct eloop *eloop, int fd,
+    void (*write_cb)(void *), void *write_cb_arg)
+{
+
+	return eloop_event_add_rw(eloop, fd, NULL,NULL, write_cb, write_cb_arg);
+}
+
 void
 eloop_event_delete_write(struct eloop *eloop, int fd, int write_only)
 {
@@ -265,52 +429,53 @@ eloop_event_delete_write(struct eloop *eloop, int fd, int write_only)
 
 	assert(eloop != NULL);
 
-	TAILQ_FOREACH(e, &eloop->events, next) {
-		if (e->fd == fd) {
-			if (write_only && e->read_cb != NULL) {
-				if (e->write_cb != NULL) {
-					e->write_cb = NULL;
-					e->write_cb_arg = NULL;
+	if (fd > eloop->events_maxfd ||
+	    (e = eloop->event_fds[fd]) == NULL)
+		return;
+
+	if (write_only) {
+		if (e->write_cb == NULL)
+			return;
+		if (e->read_cb == NULL)
+			goto remove;
+		e->write_cb = NULL;
+		e->write_cb_arg = NULL;
 #if defined(HAVE_KQUEUE)
-					EV_SET(&ke[0], (uintptr_t)fd,
-					    EVFILT_WRITE, EV_DELETE,
-					    0, 0, UPTR(NULL));
-					kevent(eloop->poll_fd, ke, 1, NULL, 0,
-					    NULL);
+		EV_SET(&ke[0], (uintptr_t)e->fd,
+		    EVFILT_WRITE, EV_DELETE, 0, 0, UPTR(NULL));
+		kevent(eloop->poll_fd, ke, 1, NULL, 0, NULL);
 #elif defined(HAVE_EPOLL)
-					memset(&epe, 0, sizeof(epe));
-					epe.data.fd = e->fd;
-					epe.data.ptr = e;
-					epe.events = EPOLLIN;
-					epoll_ctl(eloop->poll_fd,
-					    EPOLL_CTL_MOD, fd, &epe);
+		memset(&epe, 0, sizeof(epe));
+		epe.data.fd = e->fd;
+		epe.data.ptr = e;
+		epe.events = EPOLLIN;
+		epoll_ctl(eloop->poll_fd, EPOLL_CTL_MOD, fd, &epe);
 #endif
-				}
-			} else {
-				TAILQ_REMOVE(&eloop->events, e, next);
-#if defined(HAVE_KQUEUE)
-				EV_SET(&ke[0], (uintptr_t)fd, EVFILT_READ,
-				    EV_DELETE, 0, 0, UPTR(NULL));
-				if (e->write_cb)
-					EV_SET(&ke[1], (uintptr_t)fd,
-					    EVFILT_WRITE, EV_DELETE,
-					    0, 0, UPTR(NULL));
-				kevent(eloop->poll_fd, ke, e->write_cb ? 2 : 1,
-				    NULL, 0, NULL);
-#elif defined(HAVE_EPOLL)
-				/* NULL event is safe because we
-				 * rely on epoll_pwait which as added
-				 * after the delete without event was fixed. */
-				epoll_ctl(eloop->poll_fd, EPOLL_CTL_DEL,
-				    fd, NULL);
-#endif
-				TAILQ_INSERT_TAIL(&eloop->free_events, e, next);
-				eloop->events_len--;
-			}
-			eloop_event_setup_fds(eloop);
-			break;
-		}
+		eloop_event_setup_fds(eloop);
+		return;
 	}
+
+remove:
+	TAILQ_REMOVE(&eloop->events, e, next);
+	eloop->event_fds[e->fd] = NULL;
+	TAILQ_INSERT_TAIL(&eloop->free_events, e, next);
+	eloop->events_len--;
+
+#if defined(HAVE_KQUEUE)
+	EV_SET(&ke[0], (uintptr_t)fd, EVFILT_READ,
+	    EV_DELETE, 0, 0, UPTR(NULL));
+	if (e->write_cb)
+		EV_SET(&ke[1], (uintptr_t)fd,
+		    EVFILT_WRITE, EV_DELETE, 0, 0, UPTR(NULL));
+	kevent(eloop->poll_fd, ke, e->write_cb ? 2 : 1, NULL, 0, NULL);
+#elif defined(HAVE_EPOLL)
+	/* NULL event is safe because we
+	 * rely on epoll_pwait which as added
+	 * after the delete without event was fixed. */
+	epoll_ctl(eloop->poll_fd, EPOLL_CTL_DEL, fd, NULL);
+#endif
+
+	eloop_event_setup_fds(eloop);
 }
 
 int
@@ -332,7 +497,7 @@ eloop_q_timeout_add_tv(struct eloop *eloop, int queue,
 		return -1;
 	}
 
-	/* Remove existing timeout if present */
+	/* Remove existing timeout if present. */
 	TAILQ_FOREACH(t, &eloop->timeouts, next) {
 		if (t->callback == callback && t->arg == arg) {
 			TAILQ_REMOVE(&eloop->timeouts, t, next);
@@ -341,7 +506,7 @@ eloop_q_timeout_add_tv(struct eloop *eloop, int queue,
 	}
 
 	if (t == NULL) {
-		/* No existing, so allocate or grab one from the free pool */
+		/* No existing, so allocate or grab one from the free pool. */
 		if ((t = TAILQ_FIRST(&eloop->free_timeouts))) {
 			TAILQ_REMOVE(&eloop->free_timeouts, t, next);
 		} else {
@@ -448,18 +613,25 @@ eloop_open(struct eloop *eloop)
 	{
 		close(eloop->poll_fd);
 		eloop->poll_fd = -1;
-		return -1;
 	}
 
 	return eloop->poll_fd;
 #elif defined (HAVE_EPOLL)
 	return (eloop->poll_fd = epoll_create1(EPOLL_CLOEXEC));
+#else
+	return (eloop->poll_fd = -1);
 #endif
 }
+#endif
 
 int
 eloop_requeue(struct eloop *eloop)
 {
+#if defined(HAVE_POLL)
+
+	UNUSED(eloop);
+	return 0;
+#else /* !HAVE_POLL */
 	struct eloop_event *e;
 	int error;
 #if defined(HAVE_KQUEUE)
@@ -520,8 +692,8 @@ eloop_requeue(struct eloop *eloop)
 #endif
 
 	return error;
+#endif /* HAVE_POLL */
 }
-#endif
 
 int
 eloop_signal_set_cb(struct eloop *eloop,
@@ -572,20 +744,20 @@ int
 eloop_signal_mask(struct eloop *eloop, sigset_t *oldset)
 {
 	sigset_t newset;
-#ifndef HAVE_KQUEUE
 	size_t i;
+#ifndef HAVE_KQUEUE
 	struct sigaction sa;
 #endif
 
 	assert(eloop != NULL);
 
-	sigfillset(&newset);
+	sigemptyset(&newset);
+	for (i = 0; i < eloop->signals_len; i++)
+		sigaddset(&newset, eloop->signals[i]);
 	if (sigprocmask(SIG_SETMASK, &newset, oldset) == -1)
 		return -1;
 
-#ifdef HAVE_KQUEUE
-	UNUSED(eloop);
-#else
+#ifndef HAVE_KQUEUE
 	memset(&sa, 0, sizeof(sa));
 	sa.sa_sigaction = eloop_signal3;
 	sa.sa_flags = SA_SIGINFO;
@@ -612,13 +784,16 @@ eloop_new(void)
 	eloop = calloc(1, sizeof(*eloop));
 	if (eloop) {
 		TAILQ_INIT(&eloop->events);
+		eloop->events_maxfd = -1;
 		TAILQ_INIT(&eloop->free_events);
 		TAILQ_INIT(&eloop->timeouts);
 		TAILQ_INIT(&eloop->free_timeouts);
 		eloop->exitcode = EXIT_FAILURE;
 #if defined(HAVE_KQUEUE) || defined(HAVE_EPOLL)
-		eloop->poll_fd = -1;
-		eloop_open(eloop);
+		if (eloop_open(eloop) == -1) {
+			eloop_free(eloop);
+			return NULL;
+		}
 #endif
 	}
 
@@ -633,6 +808,7 @@ void eloop_free(struct eloop *eloop)
 	if (eloop == NULL)
 		return;
 
+	free(eloop->event_fds);
 	while ((e = TAILQ_FIRST(&eloop->events))) {
 		TAILQ_REMOVE(&eloop->events, e, next);
 		free(e);
@@ -651,7 +827,7 @@ void eloop_free(struct eloop *eloop)
 	}
 #if defined(HAVE_KQUEUE) || defined(HAVE_EPOLL)
 	close(eloop->poll_fd);
-#else
+#elif defined(HAVE_POLL)
 	free(eloop->fds);
 #endif
 	free(eloop);
@@ -683,7 +859,7 @@ eloop_start(struct eloop *eloop, sigset_t *signals)
 		if (eloop->exitnow)
 			break;
 
-		/* Run all timeouts first */
+		/* Run all timeouts first. */
 		if (eloop->timeout0) {
 			t0 = eloop->timeout0;
 			eloop->timeout0 = NULL;
@@ -701,7 +877,7 @@ eloop_start(struct eloop *eloop, sigset_t *signals)
 			timespecsub(&t->when, &now, &ts);
 			tsp = &ts;
 		} else
-			/* No timeouts, so wait forever */
+			/* No timeouts, so wait forever. */
 			tsp = NULL;
 
 		if (tsp == NULL && eloop->events_len == 0)
@@ -727,9 +903,9 @@ eloop_start(struct eloop *eloop, sigset_t *signals)
 			    timeout, signals);
 		else
 			n = epoll_wait(eloop->poll_fd, &epe, 1, timeout);
-#else
+#elif defined(HAVE_POLL)
 		if (signals)
-			n = pollts(eloop->fds, (nfds_t)eloop->events_len,
+			n = POLLTS(eloop->fds, (nfds_t)eloop->events_len,
 			    tsp, signals);
 		else
 			n = poll(eloop->fds, (nfds_t)eloop->events_len,
@@ -775,18 +951,24 @@ eloop_start(struct eloop *eloop, sigset_t *signals)
 				continue;
 			}
 		}
-#else
+#elif defined(HAVE_POLL)
 		if (n > 0) {
-			TAILQ_FOREACH(e, &eloop->events, next) {
-				if (e->pollfd->revents & POLLOUT &&
-				    e->write_cb != NULL)
-				{
-					e->write_cb(e->write_cb_arg);
-					break;
+			size_t i;
+
+			for (i = 0; i < eloop->events_len; i++) {
+				if (eloop->fds[i].revents & POLLOUT) {
+					e = eloop->event_fds[eloop->fds[i].fd];
+					if (e->write_cb != NULL) {
+						e->write_cb(e->write_cb_arg);
+						break;
+					}
 				}
-				if (e->pollfd->revents && e->read_cb != NULL) {
-					e->read_cb(e->read_cb_arg);
-					break;
+				if (eloop->fds[i].revents) {
+					e = eloop->event_fds[eloop->fds[i].fd];
+					if (e->read_cb != NULL) {
+						e->read_cb(e->read_cb_arg);
+						break;
+					}
 				}
 			}
 		}
