@@ -96,11 +96,12 @@
 #define O_IPV6RA_AUTOCONF	O_BASE + 38
 #define O_IPV6RA_NOAUTOCONF	O_BASE + 39
 #define O_REJECT		O_BASE + 40
-#define O_IPV6RA_ACCEPT_NOPUBLIC	O_BASE + 41
 #define O_BOOTP			O_BASE + 42
 #define O_DEFINEND		O_BASE + 43
 #define O_NODELAY		O_BASE + 44
 #define O_INFORM6		O_BASE + 45
+#define O_LASTLEASE_EXTEND	O_BASE + 46
+#define O_INACTIVE		O_BASE + 47
 
 const struct option cf_options[] = {
 	{"background",      no_argument,       NULL, 'b'},
@@ -166,7 +167,6 @@ const struct option cf_options[] = {
 	{"ipv6ra_fork",     no_argument,       NULL, O_IPV6RA_FORK},
 	{"ipv6ra_own",      no_argument,       NULL, O_IPV6RA_OWN},
 	{"ipv6ra_own_default", no_argument,    NULL, O_IPV6RA_OWN_D},
-	{"ipv6ra_accept_nopublic", no_argument, NULL, O_IPV6RA_ACCEPT_NOPUBLIC},
 	{"ipv4",            no_argument,       NULL, O_IPV4},
 	{"noipv4",          no_argument,       NULL, O_NOIPV4},
 	{"ipv6",            no_argument,       NULL, O_IPV6},
@@ -200,6 +200,8 @@ const struct option cf_options[] = {
 	{"bootp",           no_argument,       NULL, O_BOOTP},
 	{"nodelay",         no_argument,       NULL, O_NODELAY},
 	{"noup",            no_argument,       NULL, O_NOUP},
+	{"lastleaseextend", no_argument,       NULL, O_LASTLEASE_EXTEND},
+	{"inactive",        no_argument,       NULL, O_INACTIVE},
 	{NULL,              0,                 NULL, '\0'}
 };
 
@@ -285,7 +287,7 @@ parse_string_hwaddr(char *sbuf, size_t slen, const char *str, int clid)
 	size_t l;
 	const char *p;
 	int i, punt_last = 0;
-	char c[4];
+	char c[4], cmd;
 
 	/* If surrounded by quotes then it's a string */
 	if (*str == '"') {
@@ -323,28 +325,25 @@ parse_string_hwaddr(char *sbuf, size_t slen, const char *str, int clid)
 		}
 		if (*str == '\\') {
 			str++;
-			switch(*str) {
+			switch((cmd = *str++)) {
 			case '\0':
+				str--;
 				break;
 			case 'b':
 				if (sbuf)
 					*sbuf++ = '\b';
-				str++;
 				break;
 			case 'n':
 				if (sbuf)
 					*sbuf++ = '\n';
-				str++;
 				break;
 			case 'r':
 				if (sbuf)
 					*sbuf++ = '\r';
-				str++;
 				break;
 			case 't':
 				if (sbuf)
 					*sbuf++ = '\t';
-				str++;
 				break;
 			case 'x':
 				/* Grab a hex code */
@@ -378,8 +377,7 @@ parse_string_hwaddr(char *sbuf, size_t slen, const char *str, int clid)
 				break;
 			default:
 				if (sbuf)
-					*sbuf++ = *str;
-				str++;
+					*sbuf++ = cmd;
 				break;
 			}
 		} else {
@@ -429,12 +427,14 @@ parse_iaid(uint8_t *iaid, const char *arg, size_t len)
 	return parse_iaid1(iaid, arg, len, 1);
 }
 
+#ifdef AUTH
 static int
 parse_uint32(uint32_t *i, const char *arg)
 {
 
 	return parse_iaid1((uint8_t *)i, arg, sizeof(uint32_t), 0);
 }
+#endif
 
 static char **
 splitv(struct dhcpcd_ctx *ctx, int *argc, char **argv, const char *arg)
@@ -624,6 +624,7 @@ strskipwhite(const char *s)
 	return UNCONST(s);
 }
 
+#ifdef AUTH
 /* Find the end pointer of a string. */
 static char *
 strend(const char *s)
@@ -645,6 +646,7 @@ strend(const char *s)
 	}
 	return UNCONST(++s);
 }
+#endif
 
 static int
 parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
@@ -663,12 +665,14 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	struct dhcp_opt **dop, *ndop;
 	size_t *dop_len, dl, odl;
 	struct vivco *vivco;
-	struct token *token;
 	struct group *grp;
+#ifdef AUTH
+	struct token *token;
+#endif
 #ifdef _REENTRANT
 	struct group grpbuf;
 #endif
-#ifdef INET6
+#ifdef DHCP6
 	size_t sl;
 	struct if_ia *ia;
 	uint8_t iaid[4];
@@ -680,6 +684,12 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 #ifdef INET6
 	i = 0;
 #endif
+
+/* Add a guard for static analysers.
+ * This should not be needed really because of the argument_required option
+ * in the options declaration above. */
+#define ARG_REQUIRED if (arg == NULL) goto arg_required
+
 	switch(opt) {
 	case 'f': /* FALLTHROUGH */
 	case 'g': /* FALLTHROUGH */
@@ -695,6 +705,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options |= DHCPCD_BACKGROUND;
 		break;
 	case 'c':
+		ARG_REQUIRED;
 		free(ifo->script);
 		ifo->script = strdup(arg);
 		if (ifo->script == NULL)
@@ -704,6 +715,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options |= DHCPCD_DEBUG;
 		break;
 	case 'e':
+		ARG_REQUIRED;
 		add_environ(ctx, ifo, arg, 1);
 		break;
 	case 'h':
@@ -739,6 +751,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		*ifo->vendorclassid = (uint8_t)s;
 		break;
 	case 'j':
+		ARG_REQUIRED;
 		/* per interface logging is not supported
 		 * don't want to overide the commandline */
 		if (ifname == NULL && ctx->logfile == NULL) {
@@ -751,6 +764,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options |= DHCPCD_RELEASE;
 		break;
 	case 'l':
+		ARG_REQUIRED;
 		ifo->leasetime = (uint32_t)strtou(arg, NULL,
 		    0, 0, UINT32_MAX, &e);
 		if (e) {
@@ -759,6 +773,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'm':
+		ARG_REQUIRED;
 		ifo->metric = (int)strtoi(arg, NULL, 0, 0, INT32_MAX, &e);
 		if (e) {
 			logger(ctx, LOG_ERR, "failed to convert metric %s", arg);
@@ -766,6 +781,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'o':
+		ARG_REQUIRED;
 		arg = set_option_space(ctx, arg, &d, &dl, &od, &odl, ifo,
 		    &request, &require, &no, &reject);
 		if (make_option_mask(d, dl, od, odl, request, arg, 1) != 0 ||
@@ -777,6 +793,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case O_REJECT:
+		ARG_REQUIRED;
 		arg = set_option_space(ctx, arg, &d, &dl, &od, &odl, ifo,
 		    &request, &require, &no, &reject);
 		if (make_option_mask(d, dl, od, odl, reject, arg, 1) != 0 ||
@@ -815,6 +832,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options |= DHCPCD_INFORM6;
 		break;
 	case 't':
+		ARG_REQUIRED;
 		ifo->timeout = (time_t)strtoi(arg, NULL, 0, 0, INT32_MAX, &e);
 		if (e) {
 			logger(ctx, LOG_ERR, "failed to convert timeout");
@@ -835,6 +853,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'v':
+		ARG_REQUIRED;
 		p = strchr(arg, ',');
 		if (!p || !p[1]) {
 			logger(ctx, LOG_ERR, "invalid vendor format: %s", arg);
@@ -906,6 +925,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'y':
+		ARG_REQUIRED;
 		ifo->reboot = (time_t)strtoi(arg, NULL, 0, 0, UINT32_MAX, &e);
 		if (e) {
 			logger(ctx, LOG_ERR, "failed to convert reboot %s", arg);
@@ -913,6 +933,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'z':
+		ARG_REQUIRED;
 		if (ifname == NULL)
 			ctx->ifav = splitv(ctx, &ctx->ifac, ctx->ifav, arg);
 		break;
@@ -925,6 +946,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options &= ~DHCPCD_DAEMONISE;
 		break;
 	case 'C':
+		ARG_REQUIRED;
 		/* Commas to spaces for shell */
 		while ((p = strchr(arg, ',')))
 			*p = ' ';
@@ -996,6 +1018,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->options |= DHCPCD_MASTER;
 		break;
 	case 'O':
+		ARG_REQUIRED;
 		arg = set_option_space(ctx, arg, &d, &dl, &od, &odl, ifo,
 		    &request, &require, &no, &reject);
 		if (make_option_mask(d, dl, od, odl, request, arg, -1) != 0 ||
@@ -1007,6 +1030,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'Q':
+		ARG_REQUIRED;
 		arg = set_option_space(ctx, arg, &d, &dl, &od, &odl, ifo,
 		    &request, &require, &no, &reject);
 		if (make_option_mask(d, dl, od, odl, require, arg, 1) != 0 ||
@@ -1019,6 +1043,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case 'S':
+		ARG_REQUIRED;
 		p = strchr(arg, '=');
 		if (p == NULL) {
 			logger(ctx, LOG_ERR, "static assignment required");
@@ -1069,7 +1094,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 				*fp = ' ';
 				return -1;
 			}
-			if (parse_addr(ctx, &rt->dest, &rt->net, p) == -1 ||
+			if (parse_addr(ctx, &rt->dest, &rt->mask, p) == -1 ||
 			    parse_addr(ctx, &rt->gate, NULL, np) == -1)
 			{
 				free(rt);
@@ -1094,7 +1119,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 				return -1;
 			}
 			rt->dest.s_addr = INADDR_ANY;
-			rt->net.s_addr = INADDR_ANY;
+			rt->mask.s_addr = INADDR_ANY;
 			if (parse_addr(ctx, &rt->gate, NULL, p) == -1) {
 				free(rt);
 				return -1;
@@ -1110,6 +1135,24 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 				logger(ctx, LOG_ERR, "invalid MTU %s", p);
 				return -1;
 			}
+		} else if (strncmp(arg, "ip6_address=", strlen("ip6_address=")) == 0) {
+			np = strchr(p, '/');
+			if (np)
+				*np++ = '\0';
+			if (inet_pton(AF_INET6, p, &ifo->req_addr6) == 1) {
+				if (np) {
+					ifo->req_prefix_len = (uint8_t)strtou(np,
+					    NULL, 0, 0, 128, &e);
+					if (e) {
+						logger(ctx, LOG_ERR,
+						    "%s: failed to "
+						    "convert prefix len",
+						    ifname);
+						return -1;
+					}
+				} else
+					ifo->req_prefix_len = 128;
+			}
 		} else {
 			dl = 0;
 			if (ifo->config != NULL) {
@@ -1138,6 +1181,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			nconf = realloc(ifo->config, sizeof(char *) * (dl + 2));
 			if (nconf == NULL) {
 				logger(ctx, LOG_ERR, "%s: %m", __func__);
+				free(p);
 				return -1;
 			}
 			ifo->config = nconf;
@@ -1176,6 +1220,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ifo->blacklist[ifo->blacklist_len++] = addr2.s_addr;
 		break;
 	case 'Z':
+		ARG_REQUIRED;
 		if (ifname == NULL)
 			ctx->ifdv = splitv(ctx, &ctx->ifdc, ctx->ifdv, arg);
 		break;
@@ -1222,6 +1267,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case O_DESTINATION:
+		ARG_REQUIRED;
 		arg = set_option_space(ctx, arg, &d, &dl, &od, &odl, ifo,
 		    &request, &require, &no, &reject);
 		if (make_option_mask(d, dl, od, odl,
@@ -1236,6 +1282,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		}
 		break;
 	case O_FALLBACK:
+		ARG_REQUIRED;
 		free(ifo->fallback);
 		ifo->fallback = strdup(arg);
 		if (ifo->fallback == NULL) {
@@ -1245,6 +1292,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		break;
 #endif
 	case O_IAID:
+		ARG_REQUIRED;
 		if (ifname == NULL) {
 			logger(ctx, LOG_ERR,
 			    "IAID must belong in an interface block");
@@ -1271,9 +1319,6 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	case O_IPV6RA_OWN_D:
 		ifo->options |= DHCPCD_IPV6RA_OWN_DEFAULT;
 		break;
-	case O_IPV6RA_ACCEPT_NOPUBLIC:
-		ifo->options |= DHCPCD_IPV6RA_ACCEPT_NOPUBLIC;
-		break;
 	case O_IPV6RA_AUTOCONF:
 		ifo->options |= DHCPCD_IPV6RA_AUTOCONF;
 		break;
@@ -1283,7 +1328,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	case O_NOALIAS:
 		ifo->options |= DHCPCD_NOALIAS;
 		break;
-#ifdef INET6
+#ifdef DHCP6
 	case O_IA_NA:
 		i = D6_OPTION_IA_NA;
 		/* FALLTHROUGH */
@@ -1323,7 +1368,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ia = NULL;
 		for (sl = 0; sl < ifo->ia_len; sl++) {
 			if ((arg == NULL && !ifo->ia[sl].iaid_set) ||
-			    (ifo->ia[sl].iaid_set &&
+			    (arg != NULL && ifo->ia[sl].iaid_set &&
 			    ifo->ia[sl].iaid[0] == iaid[0] &&
 			    ifo->ia[sl].iaid[1] == iaid[1] &&
 			    ifo->ia[sl].iaid[2] == iaid[2] &&
@@ -1371,15 +1416,15 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 					memset(&ia->addr, 0, sizeof(ia->addr));
 				}
 				if (p && ia->ia_type == D6_OPTION_IA_PD) {
-					i = (int)strtoi(p, NULL, 0, 8, 120, &e);
+					ia->prefix_len = (uint8_t)strtou(p,
+					    NULL, 0, 8, 120, &e);
 					if (e) {
 						logger(ctx, LOG_ERR,
 						    "%s: failed to convert"
 						    " prefix len",
 						    p);
 						ia->prefix_len = 0;
-					} else
-						ia->prefix_len = (uint8_t)i;
+					}
 				}
 			}
 			ia->sla_max = 0;
@@ -1408,18 +1453,19 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			if (strlcpy(sla->ifname, p,
 			    sizeof(sla->ifname)) >= sizeof(sla->ifname))
 			{
-				logger(ctx, LOG_ERR, "%s: interface name too long",
-				    arg);
+				logger(ctx, LOG_ERR,
+				    "%s: interface name too long", arg);
 				goto err_sla;
 			}
+			sla->sla_set = 0;
+			sla->prefix_len = 0;
+			sla->suffix = 1;
 			p = np;
 			if (p) {
 				np = strchr(p, '/');
 				if (np)
 					*np++ = '\0';
-				if (*p == '\0')
-					sla->sla_set = 0;
-				else {
+				if (*p != '\0') {
 					sla->sla = (uint32_t)strtou(p, NULL,
 					    0, 0, UINT32_MAX, &e);
 					sla->sla_set = 1;
@@ -1430,31 +1476,61 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 						goto err_sla;
 					}
 				}
-				if (np) {
-					sla->prefix_len = (uint8_t)strtoi(np,
-					    NULL, 0, 0, 128, &e);
+				p = np;
+			}
+			if (p) {
+				np = strchr(p, '/');
+				if (np)
+					*np++ = '\0';
+				if (*p != '\0') {
+					sla->prefix_len = (uint8_t)strtou(p,
+				    NULL, 0, 0, 120, &e);
 					if (e) {
-						logger(ctx, LOG_ERR, "%s: failed to "
+						logger(ctx, LOG_ERR,
+						    "%s: failed to "
 						    "convert prefix len",
 						    ifname);
 						goto err_sla;
 					}
-				} else
-					sla->prefix_len = 0;
-			} else {
-				sla->sla_set = 0;
-				sla->prefix_len = 0;
+				}
+				p = np;
+			}
+			if (p) {
+				np = strchr(p, '/');
+				if (np)
+					*np = '\0';
+				if (*p != '\0') {
+					sla->suffix = (uint64_t)strtou(p, NULL,
+					    0, 0, UINT64_MAX, &e);
+					if (e) {
+						logger(ctx, LOG_ERR,
+						    "%s: failed to "
+						    "convert suffix",
+						    ifname);
+						goto err_sla;
+					}
+				}
 			}
 			/* Sanity check */
 			for (sl = 0; sl < ia->sla_len - 1; sl++) {
 				slap = &ia->sla[sl];
 				if (slap->sla_set != sla->sla_set) {
-					logger(ctx, LOG_WARNING,
+					logger(ctx, LOG_ERR,
 					    "%s: cannot mix automatic "
 					    "and fixed SLA",
 					    sla->ifname);
 					goto err_sla;
 				}
+				if (ia->prefix_len &&
+				    (sla->prefix_len == ia->prefix_len ||
+				    slap->prefix_len == ia->prefix_len))
+				{
+					logger(ctx, LOG_ERR,
+					    "%s: cannot delegte the same"
+					    "prefix length more than once",
+					    sla->ifname);
+					goto err_sla;
+				}
 				if (sla->sla_set == 0 &&
 				    strcmp(slap->ifname, sla->ifname) == 0)
 				{
@@ -1465,11 +1541,13 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 					    sla->ifname);
 					goto err_sla;
 				}
-				if (slap->sla == 0 || sla->sla == 0) {
+				if (slap->sla_set && sla->sla_set &&
+				    slap->sla == sla->sla)
+				{
 					logger(ctx, LOG_ERR, "%s: cannot"
-					    " assign multiple prefixes"
-					    " with a SLA of 0",
-					    ifname);
+					    " assign the same SLA %u"
+					    " more than once",
+					    sla->ifname, sla->sla);
 					goto err_sla;
 				}
 			}
@@ -1485,6 +1563,7 @@ err_sla:
 		ifo->options |= DHCPCD_HOSTNAME | DHCPCD_HOSTNAME_SHORT;
 		break;
 	case O_DEV:
+		ARG_REQUIRED;
 #ifdef PLUGIN_DEV
 		if (ctx->dev_load)
 			free(ctx->dev_load);
@@ -1533,6 +1612,7 @@ err_sla:
 		}
 		/* FALLTHROUGH */
 	case O_ENCAP:
+		ARG_REQUIRED;
 		if (dop == NULL) {
 			if (*ldop == NULL) {
 				logger(ctx, LOG_ERR, "encap must be after a define");
@@ -1587,7 +1667,7 @@ err_sla:
 		}
 		t = 0;
 		if (strcasecmp(arg, "request") == 0) {
-			t |= REQUEST;
+			t |= OT_REQUEST;
 			arg = strskipwhite(fp);
 			fp = strwhite(arg);
 			if (fp == NULL) {
@@ -1596,7 +1676,7 @@ err_sla:
 			}
 			*fp++ = '\0';
 		} else if (strcasecmp(arg, "norequest") == 0) {
-			t |= NOREQ;
+			t |= OT_NOREQ;
 			arg = strskipwhite(fp);
 			fp = strwhite(arg);
 			if (fp == NULL) {
@@ -1606,7 +1686,7 @@ err_sla:
 			*fp++ = '\0';
 		}
 		if (strcasecmp(arg, "optional") == 0) {
-			t |= OPTIONAL;
+			t |= OT_OPTIONAL;
 			arg = strskipwhite(fp);
 			fp = strwhite(arg);
 			if (fp == NULL) {
@@ -1617,7 +1697,7 @@ err_sla:
 			*fp++ = '\0';
 		}
 		if (strcasecmp(arg, "index") == 0) {
-			t |= INDEX;
+			t |= OT_INDEX;
 			arg = strskipwhite(fp);
 			fp = strwhite(arg);
 			if (fp == NULL) {
@@ -1627,7 +1707,7 @@ err_sla:
 			*fp++ = '\0';
 		}
 		if (strcasecmp(arg, "array") == 0) {
-			t |= ARRAY;
+			t |= OT_ARRAY;
 			arg = strskipwhite(fp);
 			fp = strwhite(arg);
 			if (fp == NULL) {
@@ -1637,68 +1717,72 @@ err_sla:
 			*fp++ = '\0';
 		}
 		if (strcasecmp(arg, "ipaddress") == 0)
-			t |= ADDRIPV4;
+			t |= OT_ADDRIPV4;
 		else if (strcasecmp(arg, "ip6address") == 0)
-			t |= ADDRIPV6;
+			t |= OT_ADDRIPV6;
 		else if (strcasecmp(arg, "string") == 0)
-			t |= STRING;
+			t |= OT_STRING;
 		else if (strcasecmp(arg, "byte") == 0)
-			t |= UINT8;
+			t |= OT_UINT8;
 		else if (strcasecmp(arg, "bitflags") == 0)
-			t |= BITFLAG;
+			t |= OT_BITFLAG;
+		else if (strcasecmp(arg, "uint8") == 0)
+			t |= OT_UINT8;
+		else if (strcasecmp(arg, "int8") == 0)
+			t |= OT_INT8;
 		else if (strcasecmp(arg, "uint16") == 0)
-			t |= UINT16;
+			t |= OT_UINT16;
 		else if (strcasecmp(arg, "int16") == 0)
-			t |= SINT16;
+			t |= OT_INT16;
 		else if (strcasecmp(arg, "uint32") == 0)
-			t |= UINT32;
+			t |= OT_UINT32;
 		else if (strcasecmp(arg, "int32") == 0)
-			t |= SINT32;
+			t |= OT_INT32;
 		else if (strcasecmp(arg, "flag") == 0)
-			t |= FLAG;
+			t |= OT_FLAG;
 		else if (strcasecmp(arg, "raw") == 0)
-			t |= STRING | RAW;
+			t |= OT_STRING | OT_RAW;
 		else if (strcasecmp(arg, "ascii") == 0)
-			t |= STRING | ASCII;
+			t |= OT_STRING | OT_ASCII;
 		else if (strcasecmp(arg, "domain") == 0)
-			t |= STRING | DOMAIN | RFC1035;
+			t |= OT_STRING | OT_DOMAIN | OT_RFC1035;
 		else if (strcasecmp(arg, "dname") == 0)
-			t |= STRING | DOMAIN;
+			t |= OT_STRING | OT_DOMAIN;
 		else if (strcasecmp(arg, "binhex") == 0)
-			t |= STRING | BINHEX;
+			t |= OT_STRING | OT_BINHEX;
 		else if (strcasecmp(arg, "embed") == 0)
-			t |= EMBED;
+			t |= OT_EMBED;
 		else if (strcasecmp(arg, "encap") == 0)
-			t |= ENCAP;
+			t |= OT_ENCAP;
 		else if (strcasecmp(arg, "rfc3361") ==0)
-			t |= STRING | RFC3361;
+			t |= OT_STRING | OT_RFC3361;
 		else if (strcasecmp(arg, "rfc3442") ==0)
-			t |= STRING | RFC3442;
+			t |= OT_STRING | OT_RFC3442;
 		else if (strcasecmp(arg, "option") == 0)
-			t |= OPTION;
+			t |= OT_OPTION;
 		else {
 			logger(ctx, LOG_ERR, "unknown type: %s", arg);
 			return -1;
 		}
-		if (l && !(t & (STRING | BINHEX))) {
+		if (l && !(t & (OT_STRING | OT_BINHEX))) {
 			logger(ctx, LOG_WARNING,
 			    "ignoring length for type `%s'", arg);
 			l = 0;
 		}
-		if (t & ARRAY && t & (STRING | BINHEX) &&
-		    !(t & (RFC1035 | DOMAIN)))
+		if (t & OT_ARRAY && t & (OT_STRING | OT_BINHEX) &&
+		    !(t & (OT_RFC1035 | OT_DOMAIN)))
 		{
 			logger(ctx, LOG_WARNING, "ignoring array for strings");
-			t &= ~ARRAY;
+			t &= ~OT_ARRAY;
 		}
-		if (t & BITFLAG) {
+		if (t & OT_BITFLAG) {
 			if (bp == NULL)
 				logger(ctx, LOG_WARNING,
 				    "missing bitflag assignment");
 		}
 		/* variable */
 		if (!fp) {
-			if (!(t & OPTION)) {
+			if (!(t & OT_OPTION)) {
 			        logger(ctx, LOG_ERR,
 				    "type %s requires a variable name", arg);
 				return -1;
@@ -1718,7 +1802,7 @@ err_sla:
 				}
 			} else {
 				np = NULL;
-				t |= RESERVED;
+				t |= OT_RESERVED;
 			}
 		}
 		if (opt != O_EMBED) {
@@ -1774,6 +1858,7 @@ err_sla:
 		}
 		break;
 	case O_VENDCLASS:
+		ARG_REQUIRED;
 		fp = strwhite(arg);
 		if (fp)
 			*fp++ = '\0';
@@ -1816,6 +1901,8 @@ err_sla:
 		vivco->data = (uint8_t *)np;
 		break;
 	case O_AUTHPROTOCOL:
+		ARG_REQUIRED;
+#ifdef AUTH
 		fp = strwhite(arg);
 		if (fp)
 			*fp++ = '\0';
@@ -1864,7 +1951,13 @@ err_sla:
 		}
 		ifo->auth.options |= DHCPCD_AUTH_SEND;
 		break;
+#else
+		logger(ctx, LOG_ERR, "no authentication support");
+		return -1;
+#endif
 	case O_AUTHTOKEN:
+		ARG_REQUIRED;
+#ifdef AUTH
 		fp = strwhite(arg);
 		if (fp == NULL) {
 			logger(ctx, LOG_ERR, "authtoken requires a realm");
@@ -1957,6 +2050,10 @@ err_sla:
 		token->key = malloc(token->key_len);
 		parse_string((char *)token->key, token->key_len, arg);
 		TAILQ_INSERT_TAIL(&ifo->auth.tokens, token, next);
+#else
+		logger(ctx, LOG_ERR, "no authentication support");
+		return -1;
+#endif
 		break;
 	case O_AUTHNOTREQUIRED:
 		ifo->auth.options &= ~DHCPCD_AUTH_REQUIRE;
@@ -1974,6 +2071,7 @@ err_sla:
 		ifo->options &= ~DHCPCD_DHCP6;
 		break;
 	case O_CONTROLGRP:
+		ARG_REQUIRED;
 #ifdef _REENTRANT
 		l = sysconf(_SC_GETGR_R_SIZE_MAX);
 		if (l == -1)
@@ -2032,6 +2130,7 @@ err_sla:
 		ifo->options &= ~DHCPCD_IF_UP;
 		break;
 	case O_SLAAC:
+		ARG_REQUIRED;
 		if (strcmp(arg, "private") == 0 ||
 		    strcmp(arg, "stableprivate") == 0 ||
 		    strcmp(arg, "stable") == 0)
@@ -2045,11 +2144,24 @@ err_sla:
 	case O_NODELAY:
 		ifo->options &= ~DHCPCD_INITIAL_DELAY;
 		break;
+	case O_LASTLEASE_EXTEND:
+		ifo->options |= DHCPCD_LASTLEASE | DHCPCD_LASTLEASE_EXTEND;
+		break;
+	case O_INACTIVE:
+		ifo->options |= DHCPCD_INACTIVE;
+		break;
 	default:
 		return 0;
 	}
 
 	return 1;
+
+#ifdef ARG_REQUIRED
+arg_required:
+	logger(ctx, LOG_ERR, "option %d requires an argument", opt);
+	return -1;
+#undef ARG_REQUIRED
+#endif
 }
 
 static int
@@ -2102,8 +2214,9 @@ static char *
 get_line(char ** __restrict buf, size_t * __restrict buflen,
     FILE * __restrict fp)
 {
-	char *p;
+	char *p, *c;
 	ssize_t bytes;
+	int quoted;
 
 	do {
 		bytes = getline(buf, buflen, fp);
@@ -2114,9 +2227,52 @@ get_line(char ** __restrict buf, size_t * __restrict buflen,
 	} while (*p == '\0' || *p == '\n' || *p == '#' || *p == ';');
 	if ((*buf)[--bytes] == '\n')
 		(*buf)[bytes] = '\0';
+
+	/* Strip embedded comments unless in a quoted string or escaped */
+	quoted = 0;
+	for (c = p; *c != '\0'; c++) {
+		if (*c == '\\') {
+			c++; /* escaped */
+			continue;
+		}
+		if (*c == '"')
+			quoted = !quoted;
+		else if (*c == '#' && !quoted) {
+			*c = '\0';
+			break;
+		}
+	}
 	return p;
 }
 
+struct if_options *
+default_config(struct dhcpcd_ctx *ctx)
+{
+	struct if_options *ifo;
+
+	/* Seed our default options */
+	if ((ifo = calloc(1, sizeof(*ifo))) == NULL) {
+		logger(ctx, LOG_ERR, "%s: %m", __func__);
+		return NULL;
+	}
+	ifo->options |= DHCPCD_IF_UP | DHCPCD_LINK | DHCPCD_INITIAL_DELAY;
+	ifo->timeout = DEFAULT_TIMEOUT;
+	ifo->reboot = DEFAULT_REBOOT;
+	ifo->metric = -1;
+	ifo->auth.options |= DHCPCD_AUTH_REQUIRE;
+#ifdef AUTH
+	TAILQ_INIT(&ifo->auth.tokens);
+#endif
+
+	/* Inherit some global defaults */
+	if (ctx->options & DHCPCD_PERSISTENT)
+		ifo->options |= DHCPCD_PERSISTENT;
+	if (ctx->options & DHCPCD_SLAACPRIVATE)
+		ifo->options |= DHCPCD_SLAACPRIVATE;
+
+	return ifo;
+}
+
 struct if_options *
 read_config(struct dhcpcd_ctx *ctx,
     const char *ifname, const char *ssid, const char *profile)
@@ -2139,13 +2295,9 @@ read_config(struct dhcpcd_ctx *ctx,
 	struct dhcp_opt *ldop, *edop;
 
 	/* Seed our default options */
-	ifo = calloc(1, sizeof(*ifo));
-	if (ifo == NULL) {
-		logger(ctx, LOG_ERR, "%s: %m", __func__);
+	if ((ifo = default_config(ctx)) == NULL)
 		return NULL;
-	}
-	ifo->options |= DHCPCD_DAEMONISE | DHCPCD_LINK | DHCPCD_INITIAL_DELAY;
-	ifo->options |= DHCPCD_IF_UP;
+	ifo->options |= DHCPCD_DAEMONISE;
 #ifdef PLUGIN_DEV
 	ifo->options |= DHCPCD_DEV;
 #endif
@@ -2158,11 +2310,6 @@ read_config(struct dhcpcd_ctx *ctx,
 	ifo->options |= DHCPCD_IPV6RA_AUTOCONF | DHCPCD_IPV6RA_REQRDNSS;
 	ifo->options |= DHCPCD_DHCP6;
 #endif
-	ifo->timeout = DEFAULT_TIMEOUT;
-	ifo->reboot = DEFAULT_REBOOT;
-	ifo->metric = -1;
-	ifo->auth.options |= DHCPCD_AUTH_REQUIRE;
-	TAILQ_INIT(&ifo->auth.tokens);
 
 	vlen = dhcp_vendor((char *)ifo->vendorclassid + 1,
 	            sizeof(ifo->vendorclassid) - 1);
@@ -2212,19 +2359,25 @@ read_config(struct dhcpcd_ctx *ctx,
 		buf = malloc(buflen);
 		if (buf == NULL) {
 			logger(ctx, LOG_ERR, "%s: %m", __func__);
+			free_options(ifo);
 			return NULL;
 		}
 		ldop = edop = NULL;
 		for (e = dhcpcd_embedded_conf; *e; e++) {
 			ol = strlen(*e) + 1;
 			if (ol > buflen) {
+				char *nbuf;
+
 				buflen = ol;
-				buf = realloc(buf, buflen);
-				if (buf == NULL) {
-					logger(ctx, LOG_ERR, "%s: %m", __func__);
+				nbuf = realloc(buf, buflen);
+				if (nbuf == NULL) {
+					logger(ctx, LOG_ERR,
+					    "%s: %m", __func__);
 					free(buf);
+					free_options(ifo);
 					return NULL;
 				}
+				buf = nbuf;
 			}
 			memcpy(buf, *e, ol);
 			line = buf;
@@ -2429,7 +2582,9 @@ free_options(struct if_options *ifo)
 	size_t i;
 	struct dhcp_opt *opt;
 	struct vivco *vo;
+#ifdef AUTH
 	struct token *token;
+#endif
 
 	if (ifo) {
 		if (ifo->environ) {
@@ -2482,6 +2637,7 @@ free_options(struct if_options *ifo)
 #endif
 		free(ifo->ia);
 
+#ifdef AUTH
 		while ((token = TAILQ_FIRST(&ifo->auth.tokens))) {
 			TAILQ_REMOVE(&ifo->auth.tokens, token, next);
 			if (token->realm_len)
@@ -2489,6 +2645,7 @@ free_options(struct if_options *ifo)
 			free(token->key);
 			free(token);
 		}
+#endif
 		free(ifo);
 	}
 }
