@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -28,21 +28,19 @@
 #ifndef IPV4LL_H
 #define IPV4LL_H
 
-#include "arp.h"
-
-extern const struct in_addr inaddr_llmask;
-extern const struct in_addr inaddr_llbcast;
-
-#define LINKLOCAL_ADDR	0xa9fe0000
-#define LINKLOCAL_MASK	IN_CLASSB_NET
-#define LINKLOCAL_BRDC	(LINKLOCAL_ADDR | ~LINKLOCAL_MASK)
+#define	LINKLOCAL_ADDR		0xa9fe0000
+#define	LINKLOCAL_MASK		IN_CLASSB_NET
+#define	LINKLOCAL_BCAST		(LINKLOCAL_ADDR | ~LINKLOCAL_MASK)
 
 #ifndef IN_LINKLOCAL
 # define IN_LINKLOCAL(addr) ((addr & IN_CLASSB_NET) == LINKLOCAL_ADDR)
 #endif
 
+#ifdef IPV4LL
+#include "arp.h"
+
 struct ipv4ll_state {
-	struct in_addr addr;
+	struct ipv4_addr *addr;
 	struct arp_state *arp;
 	unsigned int conflicts;
 	struct timespec defend;
@@ -50,13 +48,13 @@ struct ipv4ll_state {
 	uint8_t down;
 };
 
-#define IPV4LL_STATE(ifp)						       \
+#define	IPV4LL_STATE(ifp)						       \
 	((struct ipv4ll_state *)(ifp)->if_data[IF_DATA_IPV4LL])
-#define IPV4LL_CSTATE(ifp)						       \
+#define	IPV4LL_CSTATE(ifp)						       \
 	((const struct ipv4ll_state *)(ifp)->if_data[IF_DATA_IPV4LL])
-#define IPV4LL_STATE_RUNNING(ifp)					       \
-	(IPV4LL_CSTATE((ifp)) && !IPV4LL_CSTATE((ifp))->down &&	       \
-	IN_LINKLOCAL(ntohl(IPV4LL_CSTATE((ifp))->addr.s_addr)))
+#define	IPV4LL_STATE_RUNNING(ifp)					       \
+	(IPV4LL_CSTATE((ifp)) && !IPV4LL_CSTATE((ifp))->down &&		       \
+	(IPV4LL_CSTATE((ifp))->addr != NULL))
 
 struct rt* ipv4ll_subnet_route(const struct interface *);
 struct rt* ipv4ll_default_route(const struct interface *);
@@ -64,9 +62,22 @@ ssize_t ipv4ll_env(char **, const char *, const struct interface *);
 void ipv4ll_start(void *);
 void ipv4ll_claimed(void *);
 void ipv4ll_handle_failure(void *);
+#ifdef HAVE_ROUTE_METRIC
+int ipv4ll_handlert(struct dhcpcd_ctx *, int, const struct rt *);
+#else
+#define	ipv4ll_handlert(a, b, c)	(0)
+#endif
 
-#define ipv4ll_free(ifp) ipv4ll_freedrop((ifp), 0);
-#define ipv4ll_drop(ifp) ipv4ll_freedrop((ifp), 1);
+#define	ipv4ll_free(ifp) ipv4ll_freedrop((ifp), 0);
+#define	ipv4ll_drop(ifp) ipv4ll_freedrop((ifp), 1);
 void ipv4ll_freedrop(struct interface *, int);
+#else
+#define	IPV4LL_STATE_RUNNING(ifp)	(0)
+#define	ipv4ll_subnet_route(ifp)	(NULL)
+#define	ipv4ll_default_route(ifp)	(NULL)
+#define	ipv4ll_handlert(a, b, c)	(0)
+#define	ipv4ll_free(a)			{}
+#define	ipv4ll_drop(a)			{}
+#endif
 
 #endif
