@@ -189,8 +189,9 @@ ipv6nd_open(struct dhcpcd_ctx *dctx)
 	ctx = dctx->ipv6;
 	if (ctx->nd_fd != -1)
 		return ctx->nd_fd;
-	ctx->nd_fd = xsocket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6,
-	    O_NONBLOCK|O_CLOEXEC);
+#define SOCK_FLAGS	SOCK_CLOEXEC | SOCK_NONBLOCK
+	ctx->nd_fd = xsocket(PF_INET6, SOCK_RAW | SOCK_FLAGS, IPPROTO_ICMPV6);
+#undef SOCK_FLAGS
 	if (ctx->nd_fd == -1)
 		return -1;
 
@@ -217,8 +218,7 @@ ipv6nd_open(struct dhcpcd_ctx *dctx)
 	    &filt, sizeof(filt)) == -1)
 		goto eexit;
 
-	eloop_event_add(dctx->eloop, ctx->nd_fd,
-	    ipv6nd_handledata, dctx, NULL, NULL);
+	eloop_event_add(dctx->eloop, ctx->nd_fd,  ipv6nd_handledata, dctx);
 	return ctx->nd_fd;
 
 eexit:
@@ -243,7 +243,7 @@ ipv6nd_makersprobe(struct interface *ifp)
 	state->rs = calloc(1, state->rslen);
 	if (state->rs == NULL)
 		return -1;
-	rs = (struct nd_router_solicit *)(void *)state->rs;
+	rs = (void *)state->rs;
 	rs->nd_rs_type = ND_ROUTER_SOLICIT;
 	rs->nd_rs_code = 0;
 	rs->nd_rs_cksum = 0;
@@ -318,7 +318,7 @@ ipv6nd_sendrsprobe(void *arg)
 		logger(ifp->ctx, LOG_WARNING,
 		    "%s: no IPv6 Routers available", ifp->name);
 		ipv6nd_drop(ifp);
-		dhcp6_drop(ifp, "EXPIRE6");
+		dhcp6_dropnondelegates(ifp);
 	}
 }
 
@@ -551,7 +551,7 @@ ipv6nd_scriptrun(struct ra *rap)
 		{
 			hasaddress = 1;
 			if (!(ap->flags & IPV6_AF_DADCOMPLETED) &&
-			    ipv6_iffindaddr(ap->iface, &ap->addr))
+			    ipv6_iffindaddr(ap->iface, &ap->addr, IN6_IFF_TENTATIVE))
 				ap->flags |= IPV6_AF_DADCOMPLETED;
 			if ((ap->flags & IPV6_AF_DADCOMPLETED) == 0) {
 				logger(ap->iface->ctx, LOG_DEBUG,
@@ -648,9 +648,9 @@ ipv6nd_dadcallback(void *arg)
 			}
 			logger(ifp->ctx, LOG_INFO, "%s: deleting address %s",
 				ifp->name, ap->saddr);
-			if (if_deladdress6(ap) == -1 &&
+			if (if_address6(RTM_DELADDR, ap) == -1 &&
 			    errno != EADDRNOTAVAIL && errno != ENXIO)
-				logger(ifp->ctx, LOG_ERR, "if_deladdress6: %m");
+				logger(ifp->ctx, LOG_ERR, "if_address6: %m");
 			dadcounter = ap->dadcounter;
 			if (ipv6_makestableprivate(&ap->addr,
 			    &ap->prefix, ap->prefix_len,
@@ -712,23 +712,18 @@ try_script:
 	}
 }
 
+#ifndef DHCP6
+/* If DHCPv6 is compiled out, supply a shim to provide an error message
+ * if IPv6RA requests DHCPv6. */
+#undef dhcp6_start
 static int
-ipv6nd_has_public_addr(const struct interface *ifp)
+dhcp6_start(__unused struct interface *ifp, __unused enum DH6S init_state)
 {
-	const struct ra *rap;
-	const struct ipv6_addr *ia;
 
-	TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next) {
-		if (rap->iface == ifp) {
-			TAILQ_FOREACH(ia, &rap->addrs, next) {
-				if (ia->flags & IPV6_AF_AUTOCONF &&
-				    ipv6_publicaddr(ia))
-					return 1;
-			}
-		}
-	}
-	return 0;
+	errno = ENOTSUP;
+	return -1;
 }
+#endif
 
 static void
 ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
@@ -798,7 +793,7 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 		return;
 	}
 
-	if (ipv6_iffindaddr(ifp, &ctx->from.sin6_addr)) {
+	if (ipv6_iffindaddr(ifp, &ctx->from.sin6_addr, IN6_IFF_TENTATIVE)) {
 		logger(ifp->ctx, LOG_DEBUG,
 		    "%s: ignoring RA from ourself %s", ifp->name, ctx->sfrom);
 		return;
@@ -820,15 +815,10 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 		if (rap) {
 			free(rap->data);
 			rap->data_len = 0;
-			rap->no_public_warned = 0;
 		}
 		new_data = 1;
 	} else
 		new_data = 0;
-	if (new_data || ifp->options->options & DHCPCD_DEBUG)
-		logger(ifp->ctx, LOG_INFO, "%s: Router Advertisement from %s",
-		    ifp->name, ctx->sfrom);
-
 	if (rap == NULL) {
 		rap = calloc(1, sizeof(*rap));
 		if (rap == NULL) {
@@ -854,6 +844,14 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 		rap->data_len = len;
 	}
 
+	/* We could change the debug level based on new_data, but some
+	 * routers like to decrease the advertised valid and preferred times
+	 * in accordance with the own prefix times which would result in too
+	 * much needless log spam. */
+	logger(ifp->ctx, new_rap ? LOG_INFO : LOG_DEBUG,
+	    "%s: Router Advertisement from %s",
+	    ifp->name, ctx->sfrom);
+
 	clock_gettime(CLOCK_MONOTONIC, &rap->acquired);
 	rap->flags = nd_ra->nd_ra_flags_reserved;
 	rap->lifetime = ntohs(nd_ra->nd_ra_router_lifetime);
@@ -924,7 +922,7 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 
 		switch (ndo->nd_opt_type) {
 		case ND_OPT_PREFIX_INFORMATION:
-			pi = (struct nd_opt_prefix_info *)(void *)ndo;
+			pi = (void *)ndo;
 			if (pi->nd_opt_pi_len != 4) {
 				logger(ifp->ctx, new_data ? LOG_ERR : LOG_DEBUG,
 				    "%s: invalid option len for prefix",
@@ -1007,7 +1005,8 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 				 * temporary address also exists then
 				 * extend the existing one rather than
 				 * create a new one */
-				if (ipv6_iffindaddr(ifp, &ap->addr) &&
+				if (ipv6_iffindaddr(ifp, &ap->addr,
+				    IN6_IFF_NOTUSEABLE) &&
 				    ipv6_settemptime(ap, 0))
 					new_ap = 0;
 				else
@@ -1050,7 +1049,7 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 			break;
 
 		case ND_OPT_MTU:
-			mtu = (struct nd_opt_mtu *)(void *)p;
+			mtu = (void *)p;
 			mtuv = ntohl(mtu->nd_opt_mtu_mtu);
 			if (mtuv < IPV6_MMTU) {
 				logger(ifp->ctx, LOG_ERR, "%s: invalid MTU %d",
@@ -1061,7 +1060,7 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 			break;
 
 		case ND_OPT_RDNSS:
-			rdnss = (struct nd_opt_rdnss *)(void *)p;
+			rdnss = (void *)p;
 			if (rdnss->nd_opt_rdnss_lifetime &&
 			    rdnss->nd_opt_rdnss_len > 1)
 				rap->hasdns = 1;
@@ -1092,19 +1091,6 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 	if (new_rap)
 		add_router(ifp->ctx->ipv6, rap);
 
-	if (!ipv6nd_has_public_addr(rap->iface) &&
-	    !(rap->iface->options->options & DHCPCD_IPV6RA_ACCEPT_NOPUBLIC) &&
-	    (!(rap->flags & ND_RA_FLAG_MANAGED) ||
-	    !dhcp6_has_public_addr(rap->iface)))
-	{
-		logger(rap->iface->ctx,
-		    rap->no_public_warned ? LOG_DEBUG : LOG_WARNING,
-		    "%s: ignoring RA from %s"
-		    " (no public prefix, no managed address)",
-		    rap->iface->name, rap->sfrom);
-		rap->no_public_warned = 1;
-		goto handle_flag;
-	}
 	if (ifp->ctx->options & DHCPCD_TEST) {
 		script_runreason(ifp, "TEST");
 		goto handle_flag;
@@ -1117,7 +1103,7 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 	/* Find any freshly added routes, such as the subnet route.
 	 * We do this because we cannot rely on recieving the kernel
 	 * notification right now via our link socket. */
-	if_initrt6(ifp);
+	if_initrt6(ifp->ctx);
 
 	ipv6_buildroutes(ifp->ctx);
 	if (ipv6nd_scriptrun(rap))
@@ -1129,13 +1115,19 @@ ipv6nd_handlera(struct dhcpcd_ctx *dctx, struct interface *ifp,
 handle_flag:
 	if (!(ifp->options->options & DHCPCD_DHCP6))
 		goto nodhcp6;
+/* Only log a DHCPv6 start error if compiled in or debugging is enabled. */
+#ifdef DHCP6
+#define LOG_DHCP6	LOG_ERR
+#else
+#define LOG_DHCP6	LOG_DEBUG
+#endif
 	if (rap->flags & ND_RA_FLAG_MANAGED) {
 		if (new_data && dhcp6_start(ifp, DH6S_INIT) == -1)
-			logger(ifp->ctx, LOG_ERR,
+			logger(ifp->ctx, LOG_DHCP6,
 			    "dhcp6_start: %s: %m", ifp->name);
 	} else if (rap->flags & ND_RA_FLAG_OTHER) {
 		if (new_data && dhcp6_start(ifp, DH6S_INFORM) == -1)
-			logger(ifp->ctx, LOG_ERR,
+			logger(ifp->ctx, LOG_DHCP6,
 			    "dhcp6_start: %s: %m", ifp->name);
 	} else {
 		if (new_data)
@@ -1152,34 +1144,6 @@ nodhcp6:
 	ipv6nd_expirera(ifp);
 }
 
-/* Run RA's we ignored becuase they had no public addresses
- * This should only be called when DHCPv6 applies a public address */
-void
-ipv6nd_runignoredra(struct interface *ifp)
-{
-	struct ra *rap;
-
-	TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next) {
-		if (rap->iface == ifp &&
-		    !rap->expired &&
-		    rap->no_public_warned)
-		{
-			rap->no_public_warned = 0;
-			logger(rap->iface->ctx, LOG_INFO,
-			    "%s: applying ignored RA from %s",
-			    rap->iface->name, rap->sfrom);
-			if (ifp->ctx->options & DHCPCD_TEST) {
-				script_runreason(ifp, "TEST");
-				continue;
-			}
-			if (ipv6nd_scriptrun(rap))
-				return;
-			eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-			eloop_timeout_delete(ifp->ctx->eloop, NULL, rap);
-		}
-	}
-}
-
 int
 ipv6nd_hasra(const struct interface *ifp)
 {
@@ -1361,17 +1325,19 @@ ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 }
 
 void
-ipv6nd_handleifa(struct dhcpcd_ctx *ctx, int cmd, const char *ifname,
-    const struct in6_addr *addr, int flags)
+ipv6nd_handleifa(int cmd, struct ipv6_addr *addr)
 {
 	struct ra *rap;
 
-	if (ctx->ipv6 == NULL)
+	/* IPv6 init may not have happened yet if we are learning
+	 * existing addresses when dhcpcd starts. */
+	if (addr->iface->ctx->ipv6 == NULL)
 		return;
-	TAILQ_FOREACH(rap, ctx->ipv6->ra_routers, next) {
-		if (strcmp(rap->iface->name, ifname))
+
+	TAILQ_FOREACH(rap, addr->iface->ctx->ipv6->ra_routers, next) {
+		if (rap->iface != addr->iface)
 			continue;
-		ipv6_handleifa_addrs(cmd, &rap->addrs, addr, flags);
+		ipv6_handleifa_addrs(cmd, &rap->addrs, addr);
 	}
 }
 
@@ -1382,6 +1348,7 @@ ipv6nd_expirera(void *arg)
 	struct ra *rap, *ran;
 	struct timespec now, lt, expire, next;
 	uint8_t expired, valid, validone;
+	struct ipv6_addr *ia;
 
 	ifp = arg;
 	clock_gettime(CLOCK_MONOTONIC, &now);
@@ -1415,6 +1382,39 @@ ipv6nd_expirera(void *arg)
 			}
 		}
 
+		/* Not every prefix is tied to an address which
+		 * the kernel can expire, so we need to handle it ourself.
+		 * Also, some OS don't support address lifetimes (Solaris). */
+		TAILQ_FOREACH(ia, &rap->addrs, next) {
+			if (ia->prefix_vltime == ND6_INFINITE_LIFETIME ||
+			    ia->prefix_vltime == 0)
+				continue;
+			lt.tv_sec = (time_t)ia->prefix_vltime;
+			lt.tv_nsec = 0;
+			timespecadd(&ia->acquired, &lt, &expire);
+			if (timespeccmp(&now, &expire, >)) {
+				if (ia->flags & IPV6_AF_ADDED) {
+					logger(ia->iface->ctx, LOG_WARNING,
+					    "%s: expired address %s",
+					    ia->iface->name, ia->saddr);
+					if (if_address6(RTM_DELADDR, ia)== -1 &&
+					    errno != EADDRNOTAVAIL &&
+					    errno != ENXIO)
+						logger(ia->iface->ctx, LOG_ERR,
+						    "if_address6: %m");
+				}
+				ia->prefix_vltime = ia->prefix_pltime = 0;
+				ia->flags &=
+				    ~(IPV6_AF_ADDED | IPV6_AF_DADCOMPLETED);
+				expired = 1;
+			} else {
+				timespecsub(&expire, &now, &lt);
+				if (!timespecisset(&next) ||
+				    timespeccmp(&next, &lt, >))
+					next = lt;
+			}
+		}
+
 		/* XXX FixMe!
 		 * We need to extract the lifetime from each option and check
 		 * if that has expired or not.
@@ -1438,7 +1438,7 @@ ipv6nd_expirera(void *arg)
 
 	/* No valid routers? Kill any DHCPv6. */
 	if (!validone)
-		dhcp6_drop(ifp, "EXPIRE6");
+		dhcp6_dropnondelegates(ifp);
 }
 
 void
@@ -1570,7 +1570,7 @@ ipv6nd_handledata(void *arg)
 	ctx = dctx->ipv6;
 	ctx->rcvhdr.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo)) +
 	    CMSG_SPACE(sizeof(int));
-	len = recvmsg(ctx->nd_fd, &ctx->rcvhdr, 0);
+	len = recvmsg_realloc(ctx->nd_fd, &ctx->rcvhdr, 0);
 	if (len == -1) {
 		logger(dctx, LOG_ERR, "recvmsg: %m");
 		eloop_event_delete(dctx->eloop, ctx->nd_fd);
