@@ -236,20 +236,24 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 #ifdef INET
 	int dhcp, ipv4ll;
 	const struct dhcp_state *state;
+#ifdef IPV4LL
 	const struct ipv4ll_state *istate;
 #endif
+#endif
 #ifdef INET6
 	const struct dhcp6_state *d6_state;
-	int dhcp6, ra;
+	int static6, dhcp6, ra;
 #endif
 
 #ifdef INET
 	dhcp = ipv4ll = 0;
 	state = D_STATE(ifp);
+#ifdef IPV4LL
 	istate = IPV4LL_CSTATE(ifp);
 #endif
+#endif
 #ifdef INET6
-	dhcp6 = ra = 0;
+	static6 = dhcp6 = ra = 0;
 	d6_state = D6_CSTATE(ifp);
 #endif
 	if (strcmp(reason, "TEST") == 0) {
@@ -261,13 +265,17 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 			ra = 1;
 #endif
 #ifdef INET
-		else if (istate && istate->addr.s_addr != INADDR_ANY)
+#ifdef IPV4LL
+		else if (istate && istate->addr != NULL)
 			ipv4ll = 1;
+#endif
 		else
 			dhcp = 1;
 #endif
 	}
 #ifdef INET6
+	else if (strcmp(reason, "STATIC6") == 0)
+		static6 = 1;
 	else if (reason[strlen(reason) - 1] == '6')
 		dhcp6 = 1;
 	else if (strcmp(reason, "ROUTERADVERT") == 0)
@@ -283,8 +291,10 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 		/* This space left intentionally blank */
 	}
 #ifdef INET
+#ifdef IPV4LL
 	else if (strcmp(reason, "IPV4LL") == 0)
 		ipv4ll = 1;
+#endif
 	else
 		dhcp = 1;
 #endif
@@ -357,9 +367,12 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 	} else if (1 == 2 /* appease ifdefs */
 #ifdef INET
 	    || (dhcp && state && state->new)
+#ifdef IPV4LL
 	    || (ipv4ll && IPV4LL_STATE_RUNNING(ifp))
 #endif
+#endif
 #ifdef INET6
+	    || (static6 && IPV6_STATE_RUNNING(ifp))
 	    || (dhcp6 && d6_state && d6_state->new)
 	    || (ra && ipv6nd_hasra(ifp))
 #endif
@@ -405,20 +418,20 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 		ssize_t psl;
 
 		pfx_len = strlen(pfx);
-		psl = print_string(NULL, 0, ESCSTRING,
+		psl = print_string(NULL, 0, OT_ESCSTRING,
 		    (const uint8_t *)ifp->ssid, ifp->ssid_len);
 		if (psl != -1) {
 			EMALLOC(elen, pfx_len + (size_t)psl + 1);
 			memcpy(env[elen], pfx, pfx_len);
 			print_string(env[elen] + pfx_len, (size_t)psl + 1,
-			    ESCSTRING,
+			    OT_ESCSTRING,
 			    (const uint8_t *)ifp->ssid, ifp->ssid_len);
 			elen++;
 		}
 	}
 #ifdef INET
 	if (dhcp && state && state->old) {
-		n = dhcp_env(NULL, NULL, state->old, ifp);
+		n = dhcp_env(NULL, NULL, state->old, state->old_len, ifp);
 		if (n == -1)
 			goto eexit;
 		if (n > 0) {
@@ -427,7 +440,8 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 			if (nenv == NULL)
 				goto eexit;
 			env = nenv;
-			n = dhcp_env(env + elen, "old", state->old, ifp);
+			n = dhcp_env(env + elen, "old",
+			    state->old, state->old_len, ifp);
 			if (n == -1)
 				goto eexit;
 			elen += (size_t)n;
@@ -458,6 +472,7 @@ make_env(const struct interface *ifp, const char *reason, char ***argv)
 
 dumplease:
 #ifdef INET
+#ifdef IPV4LL
 	if (ipv4ll) {
 		n = ipv4ll_env(NULL, NULL, ifp);
 		if (n > 0) {
@@ -472,8 +487,9 @@ dumplease:
 			elen += (size_t)n;
 		}
 	}
+#endif
 	if (dhcp && state && state->new) {
-		n = dhcp_env(NULL, NULL, state->new, ifp);
+		n = dhcp_env(NULL, NULL, state->new, state->new_len, ifp);
 		if (n > 0) {
 			nenv = realloc(env, sizeof(char *) *
 			    (elen + (size_t)n + 1));
@@ -481,7 +497,7 @@ dumplease:
 				goto eexit;
 			env = nenv;
 			n = dhcp_env(env + elen, "new",
-			    state->new, ifp);
+			    state->new, state->new_len, ifp);
 			if (n == -1)
 				goto eexit;
 			elen += (size_t)n;
@@ -492,6 +508,20 @@ dumplease:
 	}
 #endif
 #ifdef INET6
+	if (static6) {
+		n = ipv6_env(NULL, NULL, ifp);
+		if (n > 0) {
+			nenv = realloc(env, sizeof(char *) *
+			    (elen + (size_t)n + 1));
+			if (nenv == NULL)
+				goto eexit;
+			env = nenv;
+			n = ipv6_env(env + elen, "new", ifp);
+			if (n == -1)
+				goto eexit;
+			elen += (size_t)n;
+		}
+	}
 	if (dhcp6 && D6_STATE_RUNNING(ifp)) {
 		n = dhcp6_env(NULL, NULL, ifp,
 		    d6_state->new, d6_state->new_len);
@@ -572,9 +602,9 @@ send_interface1(struct fd_list *fd, const struct interface *iface,
 	elen = (size_t)arraytostr((const char *const *)env, &s);
 	if ((ssize_t)elen == -1) {
 		free(s);
-		return -1;
-	}
-	retval = control_queue(fd, s, elen, 1);
+		retval = -1;
+	} else
+		retval = control_queue(fd, s, elen, 1);
 	ep = env;
 	while (*ep)
 		free(*ep++);
@@ -613,13 +643,19 @@ send_interface(struct fd_list *fd, const struct interface *ifp)
 		if (send_interface1(fd, ifp, d->reason) == -1)
 			retval = -1;
 	}
+#ifdef IPV4LL
 	if (IPV4LL_STATE_RUNNING(ifp)) {
 		if (send_interface1(fd, ifp, "IPV4LL") == -1)
 			retval = -1;
 	}
 #endif
+#endif
 
 #ifdef INET6
+	if (IPV6_STATE_RUNNING(ifp)) {
+		if (send_interface1(fd, ifp, "STATIC6") == -1)
+			retval = -1;
+	}
 	if (RS_STATE_RUNNING(ifp)) {
 		if (send_interface1(fd, ifp, "ROUTERADVERT") == -1)
 			retval = -1;
@@ -670,7 +706,7 @@ script_runreason(const struct interface *ifp, const char *reason)
 
 	/* Resize for PATH and RC_SVCNAME */
 	svcname = getenv(RC_SVCNAME);
-	ep = realloc(env, sizeof(char *) * (elen + 2 + (svcname ? 1 : 0)));
+	ep = reallocarray(env, elen + 2 + (svcname ? 1 : 0), sizeof(char *));
 	if (ep == NULL) {
 		elen = 0;
 		goto out;
@@ -757,7 +793,9 @@ out:
 	while (*ep)
 		free(*ep++);
 	free(env);
-	if (elen == 0)
+	if (elen == 0) {
+		logger(ifp->ctx, LOG_ERR, "%s: malloc: %m", __func__);
 		return -1;
+	}
 	return WEXITSTATUS(status);
 }
